# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module yoton.channels.channels_reqprep

Defines the channel classes for the req/rep pattern.

"""

import time
import threading

import yoton
from yoton.misc import basestring, bytes
from yoton.misc import getErrorMsg
from yoton.channels import BaseChannel, OBJECT


# For the req/rep channels to negotiate (simple load balancing)
REQREP_SEQ_REF = 2**63

# Define object to recognize errors
ERROR_OBJECT = "yoton_ERROR_HANDLING_REQUEST"


# Define exceoptions
class TimeoutError(Exception):
    pass


class CancelledError(Exception):
    pass


# # Try loading the exceptions from the concurrency framework
# # (or maybe not; it makes yoton less lightweight)
# try:
#     from concurrent.futures import TimeoutError, CancelledError
# except ImportError:
#     pass


class Future(object):
    """Future(req_channel, req, request_id)

    The Future object represents the future result of a request done at
    a yoton.ReqChannel.

    It enables:
      * checking whether the request is done.
      * getting the result or the exception raised during handling the request.
      * canceling the request (if it is not yet running)
      * registering callbacks to handle the result when it is available

    """

    def __init__(self, req_channel, req, request_id):

        # For being a Future object
        self._result = None
        self._status = 0  # 0:waiting, 1:running, 2:canceled, 3:error, 4:success
        self._callbacks = []

        # For handling req/rep
        self._req_channel = req_channel
        self._req = req
        self._request_id = request_id
        self._rep = bytes()
        self._replier = 0

        # For resending
        self._first_send_time = time.time()
        self._next_send_time = self._first_send_time + 0.5
        self._auto_cancel_timeout = 10.0

    def _send(self, msg):
        """_send(msg)

        For sending pre-request messages 'req?', 'req-'.

        """
        msg = msg.encode("utf-8")
        try:
            self._req_channel._send(msg, 0, self._request_id + REQREP_SEQ_REF)
        except IOError:
            # if self._closed, will call _send again, and catch IOerror,
            # which will result in one more call to cancel().
            self.cancel()

    def _resend_if_necessary(self):
        """_resend_if_necessary()

        Resends the pre-request message if we have not done so for the last
        0.5 second.

        This will also auto-cancel the message if it is resend over 20 times.

        """
        timetime = time.time()
        if self._status != 0:
            pass
        elif timetime > self._first_send_time + self._auto_cancel_timeout:
            self.cancel()
        elif timetime > self._next_send_time:
            self._send("req?")
            self._next_send_time = timetime + 0.5

    def set_auto_cancel_timeout(self, timeout):
        """set_auto_cancel_timeout(timeout):

        Set the timeout after which the call is automatically cancelled
        if it is not done yet. By default, this value is 10 seconds.

        If timeout is None, there is no limit to the wait time.

        """
        if timeout is None:
            timeout = 999999999999999999.0
        if timeout > 0:
            self._auto_cancel_timeout = float(timeout)
        else:
            raise ValueError("A timeout cannot be negative")

    def cancel(self):
        """cancel()

        Attempt to cancel the call. If the call is currently being executed
        and cannot be cancelled then the method will return False, otherwise
        the call will be cancelled and the method will return True.

        """

        if self._status == 1:
            # Running, cannot cancel
            return False
        elif self._status == 0:
            # Cancel now
            self._status = 2
            self._send("req-")
            for fn in self._callbacks:
                yoton.call_later(fn, 0, self)
            return True
        else:
            # Already done or canceled
            return True

    def cancelled(self):
        """cancelled()

        Return True if the call was successfully cancelled.

        """
        return self._status == 2

    def running(self):
        """running()

        Return True if the call is currently being executed and cannot be
        cancelled.

        """
        return self._status == 1

    def done(self):
        """done()

        Return True if the call was successfully cancelled or finished running.

        """
        return self._status in [2, 3, 4]

    def _wait(self, timeout):
        """_wait(timeout)

        Wait for the request to be handled for the specified amount of time.
        While waiting, the ReqChannel local event loop is called so that
        pre-request messages can be exchanged.

        """

        # No timout means a veeeery long timeout
        if timeout is None:
            timeout = 999999999999999999.0

        # Receive packages untill we receive the one we want,
        # or untill time runs out
        timestamp = time.time() + timeout
        while (self._status < 2) and (time.time() < timestamp):
            self._req_channel._process_events_local()
            time.sleep(0.01)  # 10 ms

    def result(self, timeout=None):
        """result(timeout=None)

        Return the value returned by the call. If the call hasn’t yet
        completed then this method will wait up to timeout seconds. If
        the call hasn’t completed in timeout seconds, then a TimeoutError
        will be raised. timeout can be an int or float. If timeout is not
        specified or None, there is no limit to the wait time.

        If the future is cancelled before completing then CancelledError
        will be raised.

        If the call raised, this method will raise the same exception.

        """

        # Wait
        self._wait(timeout)

        # Return or raise error
        if self._status < 2:
            raise TimeoutError("Result unavailable within the specified time.")
        elif self._status == 2:
            raise CancelledError("Result unavailable because request was cancelled.")
        elif self._status == 3:
            raise self._result
        else:
            return self._result

    def result_or_cancel(self, timeout=1.0):
        """result_or_cancel(timeout=1.0)

        Return the value returned by the call. If the call hasn’t yet
        completed then this method will wait up to timeout seconds. If
        the call hasn’t completed in timeout seconds, then the call is
        cancelled and the method will return None.

        """

        # Wait
        self._wait(timeout)

        # Return
        if self._status == 4:
            return self._result
        else:
            self.cancel()
            return None

    def exception(self, timeout=None):
        """exception(timeout)

        Return the exception raised by the call. If the call hasn’t yet
        completed then this method will wait up to timeout seconds. If
        the call hasn’t completed in timeout seconds, then a TimeoutError
        will be raised. timeout can be an int or float. If timeout is not
        specified or None, there is no limit to the wait time.

        If the future is cancelled before completing then CancelledError
        will be raised.

        If the call completed without raising, None is returned.

        """

        # Wait
        self._wait(timeout)

        # Return or raise error
        if self._status < 2:
            raise TimeoutError("Exception unavailable within the specified time.")
        elif self._status == 2:
            raise CancelledError("Exception unavailable because request was cancelled.")
        elif self._status == 3:
            return self._result
        else:
            return None  # no exception

    def add_done_callback(self, fn):
        """add_done_callback(fn)

        Attaches the callable fn to the future. fn will be called, with
        the future as its only argument, when the future is cancelled or
        finishes running.

        Added callables are called in the order that they were added. If
        the callable raises a Exception subclass, it will be logged and
        ignored. If the callable raises a BaseException subclass, the
        behavior is undefined.

        If the future has already completed or been cancelled, fn will be
        called immediately.

        """

        # Check
        if not hasattr(fn, "__call__"):
            raise ValueError("add_done_callback expects a callable.")

        # Add
        if self.done():
            yoton.call_later(fn, 0, self)
        else:
            self._callbacks.append(fn)

    def set_running_or_notify_cancel(self):
        """set_running_or_notify_cancel()

        This method should only be called by Executor implementations before
        executing the work associated with the Future and by unit tests.

        If the method returns False then the Future was cancelled, i.e.
        Future.cancel() was called and returned True.

        If the method returns True then the Future was not cancelled and
        has been put in the running state, i.e. calls to Future.running()
        will return True.

        This method can only be called once and cannot be called after
        Future.set_result() or Future.set_exception() have been called.

        """

        if self._status == 2:
            return False
        elif self._status == 0:
            self._status = 1
            return True
        else:
            raise RuntimeError(
                "set_running_or_notify_cancel should be called when in a clear state."
            )

    def set_result(self, result):
        """set_result(result)

        Sets the result of the work associated with the Future to result.
        This method should only be used by Executor implementations and
        unit tests.

        """

        # Set result if indeed in running state
        if self._status == 1:
            self._result = result
            self._status = 4
            for fn in self._callbacks:
                yoton.call_later(fn, 0, self)

    def set_exception(self, exception):
        """set_exception(exception)

        Sets the result of the work associated with the Future to the
        Exception exception. This method should only be used by Executor
        implementations and unit tests.

        """

        # Check
        if isinstance(exception, basestring):
            exception = Exception(exception)
        if not isinstance(exception, Exception):
            raise ValueError("exception must be an Exception instance.")

        # Set result if indeed in running state
        if self._status == 1:
            self._result = exception
            self._status = 3
            for fn in self._callbacks:
                yoton.call_later(fn, 0, self)


class ReqChannel(BaseChannel):
    """ReqChannel(context, slot_base)

    The request part of the request/reply messaging pattern.
    A ReqChannel instance sends request and receive the corresponding
    replies. The requests are replied by a yoton.RepChannel instance.

    This class adopts req/rep in a remote procedure call (RPC) scheme.
    The handling of the result is done using a yoton.Future object, which
    follows the approach specified in PEP 3148. Note that for the use
    of callbacks, the yoton event loop must run.

    Basic load balancing is performed by first asking all potential
    repliers whether they can handle a request. The actual request
    is then send to the first replier to respond.

    Parameters
    ----------
    context : yoton.Context instance
        The context that this channel uses to send messages in a network.
    slot_base : string
        The base slot name. The channel appends an extension to indicate
        message type and messaging pattern to create the final slot name.
        The final slot is used to connect channels at different contexts
        in a network

    Usage
    -----
    One performs a call on a virtual method of this object. The actual
    method is executed by the yoton.RepChannel instance. The method can be
    called with normal and keyword arguments, which can be (a
    combination of): None, bool, int, float, string, list, tuple, dict.

    Example
    -------
    # Fast, but process is idling when waiting for the response.
    reply = req.add(3,4).result(2.0) # Wait two seconds

    # Asynchronous processing, but no waiting.
    def reply_handler(future):
        ... # Handle reply
    future = req.add(3,4)
    future.add_done_callback(reply_handler)

    """

    # Notes on load balancing:
    #
    # Firstly, each request has an id. Which is an integer number
    # which is increased at each new request. The id is send via
    # the dest_seq. For pre-request messages an offset is added
    # to recognize these meta-messages.
    #
    # We use an approach I call pre-request. The req channel sends
    # a pre-request to all repliers (on the same slot) asking whether
    # they want to handle a request. The content is 'req?' and the
    # dest_seq is the request-id + offset.
    #
    # The repliers collects and queues all pre-requests. It will then
    # send a reply to acknowledge the first received pre-request. The
    # content is 'req!' and dest_seq is again request-id + offset.
    #
    # The replier is now in a state of waiting for the actual request.
    # It will not acknowledge pre-requests, but keeps queing them.
    #
    # Upon receiving the acknowledge, the requester sends (directed
    # at only the first replier to acknowledge) the real request.
    # The content is the real request and dest_seq is the request-id.
    # Right after this, a pre-request cancel message is sent to all
    # repliers. The content is 'req-' and dest_seq is request-id + offset.
    #
    # When a replier receives a pre-request cancel message, it will
    # remove the pre-request from the list. If this cancels the
    # request it was currently waiting for, the replier will go back
    # to its default state, and acknowledge the first next pre-request
    # in the queue.
    #
    # When the replier answers a request, it will go back to its default
    # state, and acknowledge the first next pre-request in the queue.
    # The replier tries to answer as quickly to pre-requests as possible.
    #
    # On the request channel, a dictionary of request items is maintained.
    # Each item has an attribute specifying whether a replier has
    # acknowledged it (and which one).

    def __init__(self, context, slot_base):
        BaseChannel.__init__(self, context, slot_base, OBJECT)

        # Queue with pending requests
        self._request_items = {}

        # Timeout
        self._next_recheck_time = time.time() + 0.2

        # Counter
        self._request_counter = 0

        # The req channel is always in event driven mode
        self._run_mode = 1

        # Bind signals to process the events for this channel
        # Bind to "received" signal for quick response and a timer
        # so we can resend requests if we do not receive anything.
        self.received.bind(self._process_events_local)
        self._timer = yoton.events.Timer(0.5, False)
        self._timer.bind(self._process_events_local)
        self._timer.start()

    def _messaging_patterns(self):
        return "req-rep", "rep-req"

    def __getattr__(self, name):
        if name.startswith("_"):
            return object.__getattribute__(self, name)
        try:
            return object.__getattribute__(self, name)
        except AttributeError:

            def proxy_function(*args, **kwargs):
                return self._handle_request(name, *args, **kwargs)

            return proxy_function

    def _handle_request(self, name, *args, **kwargs):
        """_handle_request(request, callback, **kwargs)

        Post a request. This creates a Future instance and stores
        it. A message is send asking any repliers to respond.

        The actual request will be send when a reply to our pre-request
        is received. This all hapens in the yoton event loop.

        """

        # Create request object
        request = name, args, kwargs

        # Check and convert request message
        bb = self.message_to_bytes(request)

        # Get new request id
        request_id = self._request_counter = self._request_counter + 1

        # Create new item for this request and store under the request id
        item = Future(self, bb, request_id)
        self._request_items[request_id] = item

        # Send pre-request (ask repliers who want to reply to a request)
        item._send("req?")

        # Return the Future instance
        return item

    def _resend_requests(self):
        """_resend_requests()

        See if we should resend our older requests. Periodically calling
        this method enables doing a request while the replier is not yet
        attached to the network.

        This also allows the Future objects to cancel themselves if it
        takes too long.

        """
        for request_id in [key for key in self._request_items.keys()]:
            item = self._request_items[request_id]
            # Remove items that are really old
            if item.cancelled():
                self._request_items.pop(request_id)
            else:
                item._resend_if_necessary()

    def _recv_item(self):
        """_recv_item()

        Receive item. If a reply is send that is an acknowledgement
        of a replier that it wants to handle our request, the
        correpsonding request is send to that replier.

        This is a kind of mini-event loop thingy that should be
        called periodically to keep things going.

        """

        # Receive package
        package = self._recv(False)
        if not package:
            return

        # Get the package reply id and sequence number
        dest_id = package._dest_id
        request_id = package._dest_seq

        # Check dest_id
        if not dest_id:
            return  # We only want messages that are directed directly at us
        elif dest_id != self._context._id:
            return  # This should not happen; context should make sure

        if request_id > REQREP_SEQ_REF:
            # We received a reply to us asking who can handle the request.
            # Get item, send actual request. We set the replier to indicate
            # that this request is being handled, and we can any further
            # acknowledgements from other repliers.
            request_id -= REQREP_SEQ_REF
            item = self._request_items.get(request_id, None)

            if item and not item._replier:
                # Status now changes to "running" canceling is not possible
                ok = item.set_running_or_notify_cancel()
                if not ok:
                    return

                # Send actual request to specific replier
                try:
                    self._send(item._req, package._source_id, request_id)
                except IOError:
                    pass  # Channel closed, will auto-cancel at item._send()
                item._replier = package._source_id  # mark as being processed

                # Send pre-request-cancel message to everyone
                item._send("req-")

        elif request_id > 0:
            # We received a reply to an actual request

            # Get item, remove from queue, set reply, return
            item = self._request_items.pop(request_id, None)
            if item:
                item._rep = package._data
                return item

    def _process_events_local(self, dummy=None):
        """_process_events_local()

        Process events only for this object. Used by _handle_now().

        """

        # Check periodically if we should resend (or clean up) old requests
        if time.time() > self._next_recheck_time:
            self._resend_requests()
            self._next_recheck_time = time.time() + 0.1

        # Process all received messages
        while self.pending:
            item = self._recv_item()
            if item:
                reply = self.message_from_bytes(item._rep)
                if (
                    isinstance(reply, tuple)
                    and len(reply) == 2
                    and reply[0] == ERROR_OBJECT
                ):
                    item.set_exception(reply[1])
                else:
                    item.set_result(reply)


class RepChannel(BaseChannel):
    """RepChannel(context, slot_base)

    The reply part of the request/reply messaging pattern.
    A RepChannel instance receives request and sends the corresponding
    replies. The requests are send from a yoton.ReqChannel instance.

    This class adopts req/rep in a remote procedure call (RPC) scheme.

    To use a RepChannel, subclass this class and implement the methods
    that need to be available. The reply should be (a combination of)
    None, bool, int, float, string, list, tuple, dict.

    This channel needs to be set to event or thread mode to function
    (in the first case yoton events need to be processed too).
    To stop handling events again, use set_mode('off').

    Parameters
    ----------
    context : yoton.Context instance
        The context that this channel uses to send messages in a network.
    slot_base : string
        The base slot name. The channel appends an extension to indicate
        message type and messaging pattern to create the final slot name.
        The final slot is used to connect channels at different contexts
        in a network

    """

    def __init__(self, context, slot_base):
        BaseChannel.__init__(self, context, slot_base, OBJECT)

        # Pending pre-requests
        self._pre_requests = []

        # Current pre-request and time that it was acknowledged
        self._pre_request = None
        self._pre_request_time = 0

        # Create thread
        self._thread = ThreadForReqChannel(self)

        # Create timer (do not start)
        self._timer = yoton.events.Timer(2.0, False)
        self._timer.bind(self._process_events_local)

        # By default, the replier is off
        self._run_mode = 0

    def _messaging_patterns(self):
        return "rep-req", "req-rep"

    # Node that setters for normal and event_driven mode are specified in
    # channels_base.py
    def set_mode(self, mode):
        """set_mode(mode)

        Set the replier to its operating mode, or turn it off.

        Modes:
          * 0 or 'off': do not process requests
          * 1 or 'event': use the yoton event loop to process requests
          * 2 or 'thread': process requests in a separate thread

        """

        if isinstance(mode, basestring):
            mode = mode.lower()

        if mode in [0, "off"]:
            self._run_mode = 0
        elif mode in [1, "event", "event-driven"]:
            self._run_mode = 1
            self.received.bind(self._process_events_local)
            self._timer.start()
        elif mode in [2, "thread", "thread-driven"]:
            self._run_mode = 2
            if not self._thread.is_alive():
                self._thread.start()
        else:
            raise ValueError("Invalid mode for ReqChannel instance.")

    def _handle_request(self, message):
        """_handle_request(message)

        This method is called for each request, and should return
        a reply. The message contains the name of the method to call,
        this function calls that method.

        """
        # Get name and args
        name, args, kwargs = message

        # Get function
        if not hasattr(self, name):
            raise RuntimeError("Method '%s' not implemented." % name)
        else:
            func = getattr(self, name)

        # Call
        return func(*args, **kwargs)

    def _acknowledge_next_pre_request(self):

        # Cancel current pre-request ourselves if it takes too long.
        # Failsafe, only for if resetting by requester fails somehow.
        if time.time() - self._pre_request_time > 10.0:
            self._pre_request = None

        # Send any pending pre requests
        if self._pre_requests and not self._pre_request:

            # Set current pre-request and its ack time
            package = self._pre_requests.pop(0)
            self._pre_request = package
            self._pre_request_time = time.time()

            # Send acknowledgement
            msg = "req!".encode("utf-8")
            try:
                self._send(msg, package._source_id, package._dest_seq)
            except IOError:
                pass  # Channel closed, nothing we can do about that
            #
            # print 'ack', self._context.id,  package._dest_seq-REQREP_SEQ_REF

    def _replier_iteration(self, package):
        """_replier_iteration()

        Do one iteration: process one request.

        """

        # Get request id
        request_id = package._dest_seq

        if request_id > REQREP_SEQ_REF:
            # Pre-request stuff

            # Remove offset
            request_id -= REQREP_SEQ_REF

            # Get action and pre request id
            action = package._data.decode("utf-8")

            # Remove pre-request from pending requests in case of both actions:
            # Cancel pending pre-request, prevent stacking of the same request.
            for prereq in [prereq for prereq in self._pre_requests]:
                if (
                    package._source_id == prereq._source_id
                    and package._dest_seq == prereq._dest_seq
                ):
                    self._pre_requests.remove(prereq)

            if action == "req-":
                # Cancel current pre-request
                if (
                    self._pre_request
                    and package._source_id == self._pre_request._source_id
                    and package._dest_seq == self._pre_request._dest_seq
                ):
                    self._pre_request = None

            elif action == "req?":
                # New pre-request
                self._pre_requests.append(package)

        else:
            # We are asked to handle an actual request

            # We can reset the state
            self._pre_request = None

            # Get request
            request = self.message_from_bytes(package._data)

            # Get reply
            try:
                reply = self._handle_request(request)
            except Exception:
                reply = ERROR_OBJECT, getErrorMsg()
                print("yoton.RepChannel: error handling request:")
                print(reply[1])

            # Send reply
            if True:
                try:
                    bb = self.message_to_bytes(reply)
                    self._send(bb, package._source_id, request_id)
                except IOError:
                    pass  # Channel is closed
                except Exception:
                    # Probably wrong type of reply returned by handle_request()
                    print("Warning: request could not be send:")
                    print(getErrorMsg())

    def _process_events_local(self, dummy=None):
        """_process_events_local()

        Called when a message (or more) has been received.

        """

        # If closed, unregister from signal and stop the timer
        if self.closed or self._run_mode != 1:
            self.received.unbind(self._process_events_local)
            self._timer.stop()

        # Iterate while we receive data
        while True:
            package = self._recv(False)
            if package:
                self._replier_iteration(package)
                self._acknowledge_next_pre_request()
            else:
                # We always enter this the last time
                self._acknowledge_next_pre_request()
                return

    def echo(self, arg1, sleep=0.0):
        """echo(arg1, sleep=0.0)

        Default procedure that can be used for testing. It returns
        a tuple (first_arg, context_id)

        """
        time.sleep(sleep)
        return arg1, hex(self._context.id)


class ThreadForReqChannel(threading.Thread):
    """ThreadForReqChannel(channel)

    Thread to run a RepChannel in threaded mode.

    """

    def __init__(self, channel):
        threading.Thread.__init__(self)

        # Check channel
        if not isinstance(channel, RepChannel):
            raise ValueError("The given channel must be a REP channel.")

        # Store channel
        self._channel = channel

        # Make deamon
        self.setDaemon(True)

    def run(self):
        """run()

        The handler's main loop.

        """

        # Get ref to channel. Remove ref from instance
        channel = self._channel
        del self._channel

        while True:

            # Stop?
            if channel.closed or channel._run_mode != 2:
                break

            # Wait for data (blocking, look Rob, without spinlocks :)
            package = channel._recv(2.0)
            if package:
                channel._replier_iteration(package)
                channel._acknowledge_next_pre_request()
            else:
                channel._acknowledge_next_pre_request()



#
# Copyright © 2014-2015 Colin Duquesnoy
# Copyright © 2009- The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides QtGui classes and functions.
"""
from . import PYQT6, PYQT5, PYSIDE2, PYSIDE6, PythonQtError


if PYQT6:
    from PyQt6 import QtGui
    from PyQt6.QtGui import *

    QFontMetrics.width = QFontMetrics.horizontalAdvance

    # Map missing/renamed methods
    QDrag.exec_ = QDrag.exec
    QGuiApplication.exec_ = QGuiApplication.exec
    QTextDocument.print_ = QTextDocument.print

    # Allow unscoped access for enums inside the QtGui module
    from .enums_compat import promote_enums

    promote_enums(QtGui)
    del QtGui
elif PYQT5:
    from PyQt5.QtGui import *
elif PYSIDE2:
    from PySide2.QtGui import *
elif PYSIDE6:
    from PySide6.QtGui import *

    QFontMetrics.width = QFontMetrics.horizontalAdvance

    # Map DeprecationWarning methods
    QDrag.exec_ = QDrag.exec
    QGuiApplication.exec_ = QGuiApplication.exec
else:
    raise PythonQtError("No Qt bindings could be found")



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module yoton.context

Defines the context class.

"""

import threading

from yoton import connection

from yoton.misc import str, split_address
from yoton.misc import UID, PackageQueue
from yoton.core import Package, BUF_MAX_LEN
from yoton.core import SLOT_CONTEXT

from yoton.connection import ConnectionCollection
from yoton.connection_tcp import TcpConnection
from yoton.connection_itc import ItcConnection


class Context(object):
    """Context(verbose=0, queue_params=None)

    A context represents a node in the network. It can connect to
    multiple other contexts (using a yoton.Connection.
    These other contexts can be in
    another process on the same machine, or on another machine
    connected via a network or the internet.

    This class represents a context that can be used by channel instances
    to communicate to other channels in the network. (Thus the name.)

    The context is the entity that queue routes the packages produced
    by the channels to the other context in the network, where
    the packages are distributed to the right channels. A context queues
    packages while it is not connected to any other context.

    If messages are send on a channel registered at this context while
    the context is not connected, the messages are stored by the
    context and will be send to the first connecting context.

    Example 1
    ---------
    # Create context and bind to a port on localhost
    context = yoton.Context()
    context.bind('localhost:11111')
    # Create a channel and send a message
    pub = yoton.PubChannel(context, 'test')
    pub.send('Hello world!')

    Example 2
    ---------
    # Create context and connect to the port on localhost
    context = yoton.Context()
    context.connect('localhost:11111')
    # Create a channel and receive a message
    sub = yoton.SubChannel(context, 'test')
    print(sub.recv() # Will print 'Hello world!'

    Queue params
    ------------
    The queue_params parameter allows one to specify the package queues
    used in the system. It is recommended to use the same parameters
    for every context in the network. The value of queue_params should
    be a 2-element tuple specifying queue size and discard mode. The
    latter can be 'old' (default) or 'new', meaning that if the queue
    is full, either the oldest or newest messages are discarted.

    """

    def __init__(self, verbose=0, queue_params=None):

        # Whether or not to write status information
        self._verbose = verbose

        # Store queue parameters
        if queue_params is None:
            queue_params = BUF_MAX_LEN, "old"
        if not (isinstance(queue_params, tuple) and len(queue_params) == 2):
            raise ValueError("queue_params should be a 2-element tuple.")
        self._queue_params = queue_params

        # Create unique key to identify this context
        self._id = UID().get_int()

        # Channels currently registered. Maps slots to channel instance.
        self._sending_channels = {}
        self._receiving_channels = {}

        # The list of connections
        self._connections = []
        self._connections_lock = threading.RLock()

        # Queue used during startup to collect packages
        # This queue is also subject to the _connections_lock
        self._startupQueue = PackageQueue(*queue_params)

        # For numbering and routing the packages
        self._send_seq = 0
        self._recv_seq = 0
        self._source_map = {}

    def close(self):
        """close()

        Close the context in a nice way, by closing all connections
        and all channels.

        Closing a connection means disconnecting two contexts. Closing
        a channel means disasociating a channel from its context.
        Unlike connections and channels, a Context instance can be reused
        after closing (although this might not always the best strategy).

        """

        # Close all connections (also the waiting connections!)
        for c in self.connections_all:
            c.close("Closed by the context.")

        # Close all channels
        self.close_channels()

    def close_channels(self):
        """close_channels()

        Close all channels associated with this context. This does
        not close the connections. See also close().

        """

        # Get all channels
        channels1 = [c for c in self._sending_channels.values()]
        channels2 = [c for c in self._receiving_channels.values()]

        # Close all channels
        for c in set(channels1 + channels2):
            c.close()

    ## Properties

    @property
    def connections_all(self):
        """Get a list of all Connection instances currently
        associated with this context, including pending connections
        (connections waiting for another end to connect).
        In addition to normal list indexing, the connections objects can be
        queried from this list using their name.
        """

        # Lock
        self._connections_lock.acquire()
        try:
            return [c for c in self._connections if c.is_alive]
        finally:
            self._connections_lock.release()

    @property
    def connections(self):
        """Get a list of the Connection instances currently
        active for this context.
        In addition to normal list indexing, the connections objects can be
        queried  from this list using their name.
        """
        # Lock
        self._connections_lock.acquire()

        try:

            # Clean up any dead connections
            copy = ConnectionCollection()
            to_remove = []
            for c in self._connections:
                if not c.is_alive:
                    to_remove.append(c)
                elif c.is_connected:
                    copy.append(c)

            # Clean
            for c in to_remove:
                self._connections.remove(c)

            # Return copy
            return copy

        finally:
            self._connections_lock.release()

    @property
    def connection_count(self):
        """Get the number of connected contexts. Can be used as a boolean
        to check if the context is connected to any other context.
        """
        return len(self.connections)

    @property
    def id(self):
        """The 8-byte UID of this context."""
        return self._id

    ## Public methods

    def bind(self, address, max_tries=1, name=""):
        """bind(address, max_tries=1, name='')

        Setup a connection with another Context, by being the host.
        This method starts a thread that waits for incoming connections.
        Error messages are printed when an attemped connect fails. the
        thread keeps trying until a successful connection is made, or until
        the connection is closed.

        Returns a Connection instance that represents the
        connection to the other context. These connection objects
        can also be obtained via the Context.connections property.

        Parameters
        ----------
        address : str
            Should be of the shape hostname:port. The port should be an
            integer number between 1024 and 2**16. If port does not
            represent a number, a valid port number is created using a
            hash function.
        max_tries : int
            The number of ports to try; starting from the given port,
            subsequent ports are tried until a free port is available.
            The final port can be obtained using the 'port' property of
            the returned Connection instance.
        name : string
            The name for the created Connection instance. It can
            be used as a key in the connections property.

        Notes on hostname
        -----------------
        The hostname can be:
          * The IP address, or the string hostname of this computer.
          * 'localhost': the connections is only visible from this computer.
            Also some low level networking layers are bypassed, which results
            in a faster connection. The other context should also connect to
            'localhost'.
          * 'publichost': the connection is visible by other computers on the
            same network. Optionally an integer index can be appended if
            the machine has multiple IP addresses (see socket.gethostbyname_ex).

        """

        # Trigger cleanup of closed connections
        self.connections

        # Split address in protocol, real hostname and port number
        protocol, hostname, port = split_address(address)

        # Based on protocol, instantiate connection class (currently only tcp)
        if False:  # protocol == 'itc':
            connection = ItcConnection(self, name)
        else:
            connection = TcpConnection(self, name)

        # Bind connection
        connection._bind(hostname, port, max_tries)

        # Save connection instance
        self._connections_lock.acquire()
        try:
            # Push packages from startup queue
            while len(self._startupQueue):
                connection._inject_package(self._startupQueue.pop())
            # Add connection object to list of connections
            self._connections.append(connection)
        finally:
            self._connections_lock.release()

        # Return Connection instance
        return connection

    def connect(self, address, timeout=1.0, name=""):
        """connect(self, address, timeout=1.0, name='')

        Setup a connection with another context, by connection to a
        hosting context. An error is raised when the connection could
        not be made.

        Returns a Connection instance that represents the
        connection to the other context. These connection objects
        can also be obtained via the Context.connections property.

        Parameters
        ----------
        address : str
            Should be of the shape hostname:port. The port should be an
            integer number between 1024 and 2**16. If port does not
            represent a number, a valid port number is created using a
            hash function.
        max_tries : int
            The number of ports to try; starting from the given port,
            subsequent ports are tried until a free port is available.
            The final port can be obtained using the 'port' property of
            the returned Connection instance.
        name : string
            The name for the created Connection instance. It can
            be used as a key in the connections property.

        Notes on hostname
        -----------------
        The hostname can be:
          * The IP address, or the string hostname of this computer.
          * 'localhost': the connection is only visible from this computer.
            Also some low level networking layers are bypassed, which results
            in a faster connection. The other context should also host as
            'localhost'.
          * 'publichost': the connection is visible by other computers on the
            same network. Optionally an integer index can be appended if
            the machine has multiple IP addresses (see socket.gethostbyname_ex).

        """

        # Trigger cleanup of closed connections
        self.connections

        # Split address in protocol, real hostname and port number
        protocol, hostname, port = split_address(address)

        # Based on protocol, instantiate connection class (currently only tcp)
        if False:  # protocol == 'itc':
            connection = ItcConnection(self, name)
        else:
            connection = TcpConnection(self, name)

        # Create new connection and connect it
        connection._connect(hostname, port, timeout)

        # Save connection instance
        self._connections_lock.acquire()
        try:
            # Push packages from startup queue
            while self._startupQueue:
                connection._inject_package(self._startupQueue.pop())
            # Add connection object to list of connections
            self._connections.append(connection)
        finally:
            self._connections_lock.release()

        # Send message in the network to signal a new connection
        bb = "NEW_CONNECTION".encode("utf-8")
        p = Package(bb, SLOT_CONTEXT, self._id, 0, 0, 0, 0)
        self._send_package(p)

        # Return Connection instance
        return connection

    def flush(self, timeout=5.0):
        """flush(timeout=5.0)

        Wait until all pending messages are send. This will flush all
        messages posted from the calling thread. However, it is not
        guaranteed that no new messages are posted from another thread.

        Raises an error when the flushing times out.

        """
        # Flush all connections
        for c in self.connections:
            c.flush(timeout)

        # Done (backward compatibility)
        return True

    ## Private methods used by the Channel classes

    def _register_sending_channel(self, channel, slot, slotname=""):
        """_register_sending_channel(channel, slot, slotname='')

        The channel objects use this method to register themselves
        at a particular slot.

        """

        # Check if this slot is free
        if slot in self._sending_channels:
            raise ValueError("Slot not free: " + str(slotname))

        # Register
        self._sending_channels[slot] = channel

    def _register_receiving_channel(self, channel, slot, slotname=""):
        """_register_receiving_channel(channel, slot, slotname='')

        The channel objects use this method to register themselves
        at a particular slot.

        """

        # Check if this slot is free
        if slot in self._receiving_channels:
            raise ValueError("Slot not free: " + str(slotname))

        # Register
        self._receiving_channels[slot] = channel

    def _unregister_channel(self, channel):
        """_unregister_channel(channel)

        Unregisters the given channel. That channel can no longer
        receive messages, and should no longer send messages.

        """
        for D in [self._receiving_channels, self._sending_channels]:
            for key in [key for key in D.keys()]:
                if D[key] == channel:
                    D.pop(key)

    ## Private methods to pass packages between context and io-threads

    def _send_package(self, package):
        """_send_package(package)

        Used by the channels to send a package into the network.
        This method routes the package to all currentlt connected
        connections. If there are none, the packages is queued at
        the context.

        """

        # Add number
        self._send_seq += 1
        package._source_seq = self._send_seq

        # Send to all connections, or queue if there are none
        self._connections_lock.acquire()
        try:
            ok = False
            for c in self._connections:
                if c.is_alive:  # Waiting or connected
                    c._send_package(package)
                    ok = True
            # Should we queue the package?
            if not ok:
                self._startupQueue.push(package)
        finally:
            self._connections_lock.release()

    def _recv_package(self, package, connection):
        """_recv_package(package, connection)

        Used by the connections to receive a package at this
        context. The package is distributed to all connections
        except the calling one. The package is also distributed
        to the right channel (if applicable).

        """

        # Get slot
        slot = package._slot

        # Init what to do with the package
        send_further = False
        deposit_here = False

        # Get what to do with the package
        last_seq = self._source_map.get(package._source_id, 0)
        if last_seq < package._source_seq:
            # Update source map
            self._source_map[package._source_id] = package._source_seq
            if package._dest_id == 0:
                # Add to both lists, first attach seq nr
                self._recv_seq += 1
                package._recv_seq = self._recv_seq
                send_further, deposit_here = True, True
            elif package._dest_id == self._id:
                # Add only to process list, first attach seq nr
                self._recv_seq += 1
                package._recv_seq = self._recv_seq
                deposit_here = True
            else:
                # Send package to connected nodes
                send_further = True

        # Send package to other context (over all alive connections)
        if send_further:
            self._connections_lock.acquire()
            try:
                for c in self._connections:
                    if c is connection or not c.is_alive:
                        continue
                    c._send_package(package)
            finally:
                self._connections_lock.release()

        # Process package here or pass to channel
        if deposit_here:
            if slot == SLOT_CONTEXT:
                # Context-to-context messaging;
                # A slot starting with a space reprsents the context
                self._recv_context_package(package)
            else:
                # Give package to a channel (if applicable)
                channel = self._receiving_channels.get(slot, None)
                if channel is not None:
                    channel._recv_package(package)

    def _recv_context_package(self, package):
        """_recv_context_package(package)

        Process a package addressed at the context itself. This is how
        the context handles higher-level connection tasks.

        """

        # Get message: context messages are always utf-8 encoded strings
        message = package._data.decode("utf-8")

        if message == "CLOSE_CONNECTION":
            # Close the connection. Check which one of our connections is
            # connected with the context that send this message.
            self._connections_lock.acquire()
            try:
                for c in self.connections:
                    if c.is_connected and c.id2 == package._source_id:
                        c.close(connection.STOP_CLOSED_FROM_THERE, False)
            finally:
                self._connections_lock.release()

        elif message == "NEW_CONNECTION":
            # Resend all status channels
            for channel in self._sending_channels.values():
                if hasattr(channel, "_current_message") and hasattr(
                    channel, "send_last"
                ):
                    channel.send_last()

        else:
            print("Yoton: Received unknown context message: " + message)



# -*- coding: utf-8 -*-

## Connect two context

import yoton
import time

# Context 1
ct1 = yoton.Context()
pub1 = yoton.StateChannel(ct1, "state A")
pub1.send("READY")

# Context 2
ct2 = yoton.Context()
pub2 = yoton.StateChannel(ct2, "state A")

# Context 3
ct3 = yoton.Context()
sub1 = yoton.StateChannel(ct3, "state A")

# Connect
ct1.bind("localhost:test1")
ct2.connect("localhost:test1")
ct2.bind("localhost:test2")
ct3.connect("localhost:test2")

# Get status (but wait first)
time.sleep(0.3)
print(sub1.recv())

# New status
pub2.send("READY2")
time.sleep(0.3)
print(sub1.recv())

# And back to first
pub1.send("READY")
time.sleep(0.3)
print(sub1.recv())


## Now attach another context
# Upon connecting, the connecting context will issue a context-to-context
# message to indicate a new connection. All contexts will then call
# the send_last() methods of their state channels.

# Context 2
ct4 = yoton.Context()
sub2 = yoton.StateChannel(ct4, "state A")

# Connect
ct2.bind("localhost:test3")
ct4.connect("localhost:test3")

# Get status (but wait first)
time.sleep(0.3)
print(sub2.recv())

# Ask status again (simply gives last message)
print(sub2.recv())


## Using received signals


def on_new_state(channel):
    state = channel.recv()
    print("%i received state %s" % (id(channel), state))
    if state == "stop":
        yoton.stop_event_loop()


# Bind
sub1.received.bind(on_new_state)
sub2.received.bind(on_new_state)

# Have some calls made
yoton.call_later(pub1.send, 1.0, "hello")
yoton.call_later(pub1.send, 1.5, "there")
yoton.call_later(pub1.send, 2.0, "now")
yoton.call_later(pub1.send, 2.5, "stop")

# Go!
yoton.start_event_loop()



# -*- coding: utf-8 -*-
# Copyright (C) 2013 Almar Klein

"""
This module defines file system proxies to be used for the file browser.
For now, there is only one: the native file system. But in time,
we may add proxies for ftp, S3, remote computing, etc.

This may seem like an awkward way to use the file system, but (with
small modifications) this approach can probably be used also for
opening/saving files to any file system that we implement here. This
will make Pyzo truly powerful for use in remote computing.

"""

import time
import threading
from queue import Queue, Empty
import os.path as op

from . import QtCore
from .utils import isdir


class Task:
    """Task(**params)

    A task object. Accepts params as keyword arguments.
    When overloading, dont forget to set __slots__.

    Overload and implement the 'process' method to create a task.
    Then use pushTask on a pathProxy object. Use the 'result' method to
    obtain the result (or raise an error).
    """

    __slots__ = ["_params", "_result", "_error"]

    def __init__(self, **params):
        if not params:
            params = None
        self._params = params
        self._result = None
        self._error = None

    def process(self, proxy, **params):
        """process(pathProxy, **params):
        This is the method that represents the task. Overload this to make
        the task do what is intended.
        """
        pass

    def _run(self, proxy):
        """Run the task. Don't overload or use this."""
        try:
            params = self._params or {}
            self._result = self.process(proxy, **params)
        except Exception as err:
            self._error = "Task failed: {}:\n{}".format(self, str(err))
            print(self._error)

    def result(self):
        """Get the result. Raises an error if the task failed."""
        if self._error:
            raise Exception(self._error)
        else:
            return self._result


## Proxy classes for directories and files


class PathProxy(QtCore.QObject):
    """Proxy base class for DirProxy and FileProxy.

    A proxy object is used to get information on a path (folder
    contents, or file modification time), and keep being updated about
    changes in that information.

    One uses an object by connecting to the 'changed' or 'deleted' signal.
    Use setActive(True) to receive updates on these signals. If the proxy
    is no longer needed, use close() to unregister it.

    """

    changed = QtCore.Signal()
    deleted = QtCore.Signal()
    errored = QtCore.Signal(str)  # Or should we pass an error per 'action'?

    taskFinished = QtCore.Signal(Task)

    def __init__(self, fsProxy, path):
        QtCore.QObject.__init__(self)
        self._lock = threading.RLock()
        self._fsProxy = fsProxy
        self._path = path
        self._cancelled = False
        # For tasks
        self._pendingTasks = []
        self._finishedTasks = []

    def __repr__(self):
        return '<{} "{}">'.format(self.__class__.__name__, self._path)

    def path(self):
        """Get the path of this proxy."""
        return self._path

    def track(self):
        """Start tracking this proxy object in the idle time of the
        FSProxy thread.
        """
        self._fsProxy._track(self)

    def push(self):
        """Process this proxy object asap; the object is put in the queue
        of the FSProxy, so it is updated as fast as possible.
        """
        self._cancelled = False
        self._fsProxy._push(self)

    def cancel(self):
        """Stop tracking this proxy object. Cancel processing if this
        object was in the queue.
        """
        self._fsProxy._unTrack(self)
        self._cancelled = True

    def pushTask(self, task):
        """pushTask(task)
        Give a task to the proxy to be executed in the FSProxy
        thread. The taskFinished signal will be emitted with the given
        task when it is done.
        """
        shouldPush = False
        with self._lock:
            if not self._pendingTasks:
                shouldPush = True
            self._pendingTasks.append(task)
        if shouldPush:
            self.push()

    def _processTasks(self):
        # Get pending tasks
        with self._lock:
            pendingTasks = self._pendingTasks
            self._pendingTasks = []
        # Process pending tasks
        finishedTasks = []
        for task in pendingTasks:
            task._run(self)
            finishedTasks.append(task)
        # Emit signal if there are finished tasks
        for task in finishedTasks:
            self.taskFinished.emit(task)


class DirProxy(PathProxy):
    """Proxy object for a directory. Obtain an instance of this class
    using filesystemProx.dir()
    """

    def __init__(self, *args):
        PathProxy.__init__(self, *args)
        self._dirs = set()
        self._files = set()

    def dirs(self):
        with self._lock:
            return set(self._dirs)

    def files(self):
        with self._lock:
            return set(self._files)

    def _process(self, forceUpdate=False):
        # Get info
        dirs = self._fsProxy.listDirs(self._path)
        files = self._fsProxy.listFiles(self._path)
        # Is it deleted?
        if dirs is None or files is None:
            self.deleted.emit()
            return
        # All seems ok. Update if necessary
        dirs, files = set(dirs), set(files)
        if (dirs != self._dirs) or (files != self._files):
            with self._lock:
                self._dirs, self._files = dirs, files
            self.changed.emit()
        elif forceUpdate:
            self.changed.emit()


class FileProxy(PathProxy):
    """Proxy object for a file. Obtain an instance of this class
    using filesystemProx.dir()
    """

    def __init__(self, *args):
        PathProxy.__init__(self, *args)
        self._modified = 0

    def modified(self):
        with self._lock:
            return self._modified

    def _process(self, forceUpdate=False):
        # Get info
        modified = self._fsProxy.modified(self._path)
        # Is it deleted?
        if modified is None:
            self.deleted.emit()
            return
        # All seems ok. Update if necessary
        if modified != self._modified:
            with self._lock:
                self._modified = modified
            self.changed.emit()
        elif forceUpdate:
            self.changed.emit()

    def read(self):
        pass  # ?

    def save(self):
        pass  # ?


## Proxy classes for the file system


class BaseFSProxy(threading.Thread):
    """Abstract base class for file system proxies.

    The file system proxy defines an interface that subclasses can implement
    to "become" a usable file system proxy.

    This class implements the polling of information for the DirProxy
    and FileProxy objects, and keeping them up-to-date. For this purpose
    it keeps a set of PathProxy instances that are polled when idle.
    There is also a queue for items that need processing asap. This is
    where objects are put in when they are activated.

    This class has methods to use the file system (list files and
    directories, etc.). These can be used directly, but may be slow.
    Therefor it is recommended to use the FileProxy and DirProxy objects
    instead.

    """

    # Define how often the registered dirs and files are checked
    IDLE_TIMEOUT = 0.1

    # For testing to induce extra delay. Should normally be close to zero,
    # but not exactly zero!
    IDLE_DELAY = 0.01
    QUEUE_DELAY = 0.01  # 0.5

    def __init__(self):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        #
        self._interrupt = False
        self._exit = False
        #
        self._lock = threading.RLock()
        self._q = Queue()
        self._pathProxies = set()
        #
        self.start()

    def _track(self, pathProxy):
        # todo: use weak references
        with self._lock:
            self._pathProxies.add(pathProxy)

    def _unTrack(self, pathProxy):
        with self._lock:
            self._pathProxies.discard(pathProxy)

    def _push(self, pathProxy):
        # todo: use weak ref here too?
        self._q.put(pathProxy)
        self._interrupt = True

    def stop(self, *, timeout=1.0):
        with self._lock:
            self._exit = True
            self._interrupt = True
            self._pathProxies.clear()
        self.join(timeout)

    def dir(self, path):
        """Convenience function to create a new DirProxy object."""
        return DirProxy(self, path)

    def file(self, path):
        """Convenience function to create a new FileProxy object."""
        return FileProxy(self, path)

    def run(self):

        try:
            try:
                self._run()
            except Exception as err:
                if Empty is None or self._lock is None:
                    pass  # Shutting down ...
                else:
                    print("Exception in proxy thread: " + str(err))

        except Exception:
            pass  # Interpreter is shutting down

    def _run(self):

        last_sleep = time.time()

        while True:

            # Check and reset
            self._interrupt = False
            if self._exit:
                return

            # Sleep
            now = time.time()
            if now - last_sleep > 0.1:
                last_sleep = now
                time.sleep(0.05)

            try:
                # Process items from the queue
                item = self._q.get(True, self.IDLE_TIMEOUT)
                if item is not None and not item._cancelled:
                    self._processItem(item, True)
            except Empty:
                # Queue empty, check items periodically
                self._idle()

    def _idle(self):
        # Make a copy of the set if item
        with self._lock:
            items = set(self._pathProxies)
        # Process them
        for item in items:
            if self._interrupt:
                return
            self._processItem(item)

    def _processItem(self, pathProxy, forceUpdate=False):

        # Slow down a bit
        if forceUpdate:
            time.sleep(self.QUEUE_DELAY)
        else:
            time.sleep(self.IDLE_DELAY)

        # Process
        try:
            pathProxy._process(forceUpdate)
        except Exception as err:
            pathProxy.errored.emit(str(err))

        # Process tasks
        pathProxy._processTasks()

    # To overload ...

    def listDirs(self, path):
        raise NotImplementedError()  # Should rerurn None if it does not exist

    def listFiles(self, path):
        raise NotImplementedError()  # Should rerurn None if it does not exist

    def modified(self, path):
        raise NotImplementedError()  # Should rerurn None if it does not exist

    def fileSize(self, path):
        raise NotImplementedError()  # Should rerurn None if it does not exist

    def read(self, path):
        raise NotImplementedError()  # Should rerurn None if it does not exist

    def write(self, path, bb):
        raise NotImplementedError()

    def rename(self, path):
        raise NotImplementedError()

    def remove(self, path):
        raise NotImplementedError()

    def createDir(self, path):
        raise NotImplementedError()


import os


class NativeFSProxy(BaseFSProxy):
    """File system proxy for the native file system."""

    def listDirs(self, path):
        if isdir(path):
            pp = [op.join(path, p) for p in os.listdir(path)]
            return [str(p) for p in pp if isdir(p)]

    def listFiles(self, path):
        if isdir(path):
            pp = [op.join(path, p) for p in os.listdir(path)]
            return [str(p) for p in pp if op.isfile(p)]

    def modified(self, path):
        if op.isfile(path):
            return op.getmtime(path)

    def fileSize(self, path):
        if op.isfile(path):
            return op.getsize(path)

    def read(self, path):
        if op.isfile(path):
            return open(path, "rb").read()

    def write(self, path, bb):
        with open(path, "wb") as f:
            f.write(bb)

    def rename(self, path1, path2):
        os.rename(path1, path2)

    def remove(self, path):
        if op.isfile(path):
            os.remove(path)
        elif isdir(path):
            os.rmdir(path)

    def createDir(self, path):
        if not isdir(path):
            os.makedirs(path)



"""
Tools to install miniconda from pyzo and register that env in pyzo's
shell config.
"""

import os
import sys
import stat
import time
import struct
import shutil
import threading
import subprocess
import urllib.request

import pyzo
from pyzo.qt import QtCore, QtWidgets
from pyzo import translate

base_url = "http://repo.continuum.io/miniconda/"
links = {
    "win32": "Miniconda3-latest-Windows-x86.exe",
    "win64": "Miniconda3-latest-Windows-x86_64.exe",
    "osx64": "Miniconda3-latest-MacOSX-x86_64.sh",
    "linux32": "Miniconda3-latest-Linux-x86.sh",
    "linux64": "Miniconda3-latest-Linux-x86_64.sh",
    "arm": "Miniconda3-latest-Linux-armv7l.sh",  # raspberry pi
}


# Get where we want to put miniconda installer
miniconda_path = os.path.join(pyzo.appDataDir, "miniconda")
miniconda_path += ".exe" if sys.platform.startswith("win") else ".sh"

# Get default dir where we want the env
# default_conda_dir = os.path.join(pyzo.appDataDir, 'conda_root')
default_conda_dir = (
    "C:\\miniconda3"
    if sys.platform.startswith("win")
    else os.path.expanduser("~/miniconda3")
)


def check_for_conda_env(parent=None):
    """Check if it is reasonable to ask to install a conda env. If
    users says yes, do it. If user says no, don't, and remember.
    """

    # Interested previously?
    if getattr(pyzo.config.state, "did_not_want_conda_env", False):
        print("User has previously indicated to have no interest in a conda env")
        return

    # Needed?
    if pyzo.config.shellConfigs2:
        exe = pyzo.config.shellConfigs2[0]["exe"]
        r = ""
        try:
            r = subprocess.check_output(
                [exe, "-m", "conda", "info"], shell=sys.platform.startswith("win")
            )
            r = r.decode()
        except Exception:
            pass  # no Python or no conda
        if r and "is foreign system : False" in r:
            print("First shell config looks like a conda env.")
            return

    # Ask if interested now?
    d = AskToInstallConda(parent)
    d.exec_()
    if not d.result():
        pyzo.config.state.did_not_want_conda_env = True  # Mark for next time
        return

    # Launch installer
    d = Installer(parent)
    d.exec_()


class AskToInstallConda(QtWidgets.QDialog):
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.setWindowTitle("Install a conda env?")
        self.setModal(True)

        text = "Pyzo is only an editor. To execute code, you need a Python environment.\n\n"
        text += "Do you want Pyzo to install a Python environment (miniconda)?\n"
        text += "If not, you must arrange for a Python interpreter yourself"
        if not sys.platform.startswith("win"):
            text += " or use the system Python"
        text += "."
        text += "\n(You can always launch the installer from the shell menu.)"

        self._label = QtWidgets.QLabel(text, self)
        self._no = QtWidgets.QPushButton("No thanks (dont ask again)")
        self._yes = QtWidgets.QPushButton("Yes, please install Python!")

        self._no.clicked.connect(self.reject)
        self._yes.clicked.connect(self.accept)

        vbox = QtWidgets.QVBoxLayout(self)
        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(vbox)
        vbox.addWidget(self._label, 1)
        vbox.addLayout(hbox, 0)
        hbox.addWidget(self._no, 2)
        hbox.addWidget(self._yes, 2)

        self._yes.setDefault(1)


class Installer(QtWidgets.QDialog):

    lineFromStdOut = QtCore.Signal(str)

    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.setWindowTitle("Install miniconda")
        self.setModal(True)
        self.resize(500, 500)

        text = translate(
            "bootstrapconda",
            "This will download and install miniconda on your computer.",
        )

        self._label = QtWidgets.QLabel(text, self)

        self._scipystack = QtWidgets.QCheckBox(
            translate("bootstrapconda", "Also install scientific packages"), self
        )
        self._scipystack.setChecked(True)
        self._path = QtWidgets.QLineEdit(default_conda_dir, self)
        self._progress = QtWidgets.QProgressBar(self)
        self._outputLine = QtWidgets.QLabel(self)
        self._output = QtWidgets.QPlainTextEdit(self)
        self._output.setReadOnly(True)
        self._button = QtWidgets.QPushButton("Install", self)

        self._outputLine.setSizePolicy(
            QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed
        )

        vbox = QtWidgets.QVBoxLayout(self)
        self.setLayout(vbox)
        vbox.addWidget(self._label, 0)
        vbox.addWidget(self._path, 0)
        vbox.addWidget(self._scipystack, 0)
        vbox.addWidget(self._progress, 0)
        vbox.addWidget(self._outputLine, 0)
        vbox.addWidget(self._output, 1)
        vbox.addWidget(self._button, 0)

        self._button.clicked.connect(self.go)

        self.addOutput(translate("bootstrapconda", "Waiting to start installation.\n"))
        self._progress.setVisible(False)

        self.lineFromStdOut.connect(self.setStatus)

    def setStatus(self, line):
        self._outputLine.setText(line)

    def addOutput(self, text):
        # self._output.setPlainText(self._output.toPlainText() + '\n' + text)
        cursor = self._output.textCursor()
        cursor.movePosition(cursor.End, cursor.MoveAnchor)
        cursor.insertText(text)
        cursor.movePosition(cursor.End, cursor.MoveAnchor)
        self._output.setTextCursor(cursor)
        self._output.ensureCursorVisible()

    def addStatus(self, line):
        self.addOutput("\n" + line)
        self.setStatus(line)

    def go(self):

        # Check if we can install
        try:
            self._conda_dir = self._path.text()
            if not os.path.isabs(self._conda_dir):
                raise ValueError("Given installation path must be absolute.")
            if os.path.exists(self._conda_dir):
                raise ValueError("The given installation path already exists.")
        except Exception as err:
            self.addOutput("\nCould not install:\n" + str(err))
            return

        ok = False

        try:

            # Disable user input, get ready for installation
            self._progress.setVisible(True)
            self._button.clicked.disconnect()
            self._button.setEnabled(False)
            self._scipystack.setEnabled(False)
            self._path.setEnabled(False)

            if not os.path.exists(self._conda_dir):
                self.addStatus("Downloading installer ... ")
                self._progress.setMaximum(100)
                self.download()
                self.addStatus("Done downloading installer.")
                self.make_done()

                self.addStatus("Installing (this can take a minute) ... ")
                self._progress.setMaximum(0)
                ret = self.install()
                self.addStatus(("Failed" if ret else "Done") + " installing.")
                self.make_done()

            self.post_install()

            if self._scipystack.isChecked():
                self.addStatus("Installing scientific packages ... ")
                self._progress.setMaximum(0)
                ret = self.install_scipy()
                self.addStatus("Done installing scientific packages")
                self.make_done()

            self.addStatus("Verifying ... ")
            self._progress.setMaximum(100)
            ret = self.verify()
            if ret:
                self.addOutput("Error\n" + ret)
                self.addStatus("Verification Failed!")
            else:
                self.addOutput("Done verifying")
                self.addStatus("Ready to go!")
                self.make_done()
                ok = True

        except Exception as err:
            self.addStatus("Installation failed ...")
            self.addOutput("\n\nException!\n" + str(err))

        if not ok:
            self.addOutput("\n\nWe recommend installing miniconda or anaconda, ")
            self.addOutput("and making Pyzo aware if it via the shell configuration.")
        else:
            self.addOutput(
                '\n\nYou can install additional packages by running "conda install" in the shell.'
            )

        # Wrap up, allow user to exit
        self._progress.hide()
        self._button.setEnabled(True)
        self._button.setText("Close")
        self._button.clicked.connect(self.close)

    def make_done(self):
        self._progress.setMaximum(100)
        self._progress.setValue(100)
        etime = time.time() + 0.2
        while time.time() < etime:
            time.sleep(0.01)
            QtWidgets.qApp.processEvents()

    def download(self):

        # Installer already downloaded?
        if os.path.isfile(miniconda_path):
            self.addOutput("Already downloaded.")
            return  # os.remove(miniconda_path)

        # Get url key
        key = ""
        if sys.platform.startswith("win"):
            key = "win"
        elif sys.platform.startswith("darwin"):
            key = "osx"
        elif sys.platform.startswith("linux"):
            key = "linux"
        key += "64" if is_64bit() else "32"

        # Get url
        if key not in links:
            raise RuntimeError("Cannot download miniconda for this platform.")
        url = base_url + links[key]

        _fetch_file(url, miniconda_path, self._progress)

    def install(self):
        dest = self._conda_dir

        # Clear dir
        assert not os.path.isdir(dest), "Miniconda dir already exists"
        assert " " not in dest, "miniconda dest path must not contain spaces"

        if sys.platform.startswith("win"):
            return self._run_process([miniconda_path, "/S", "/D=%s" % dest])
        else:
            os.chmod(miniconda_path, os.stat(miniconda_path).st_mode | stat.S_IEXEC)
            return self._run_process([miniconda_path, "-b", "-p", dest])

    def post_install(self):

        exe = py_exe(self._conda_dir)

        # Add Pyzo channel
        cmd = [exe, "-m", "conda", "config", "--system", "--add", "channels", "pyzo"]
        subprocess.check_call(cmd, shell=sys.platform.startswith("win"))
        self.addStatus("Added Pyzo channel to conda env")

        # Add to pyzo shell config
        if pyzo.config.shellConfigs2 and pyzo.config.shellConfigs2[0]["exe"] == exe:
            pass
        else:
            s = pyzo.ssdf.new()
            s.name = "Py3-conda"
            s.exe = exe
            s.gui = "PyQt4"
            pyzo.config.shellConfigs2.insert(0, s)
            pyzo.saveConfig()
            self.addStatus("Prepended new env to Pyzo shell configs.")

    def install_scipy(self):

        packages = [
            "numpy",
            "scipy",
            "pandas",
            "matplotlib",
            "sympy",
            #'scikit-image', 'scikit-learn',
            "pyopengl",  # 'visvis', 'imageio',
            "tornado",
            "pyqt",  #'ipython', 'jupyter',
            #'requests', 'pygments','pytest',
        ]
        exe = py_exe(self._conda_dir)
        cmd = [exe, "-m", "conda", "install", "--yes"] + packages
        return self._run_process(cmd)

    def _run_process(self, cmd):
        """Run command in a separate process, catch stdout, show lines
        in the output label. On fail, show all output in output text.
        """
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=sys.platform.startswith("win"),
        )
        catcher = StreamCatcher(p.stdout, self.lineFromStdOut)

        while p.poll() is None:
            time.sleep(0.01)
            QtWidgets.qApp.processEvents()

        catcher.join()
        if p.poll():
            self.addOutput(catcher.output())
        return p.poll()

    def verify(self):

        self._progress.setValue(1)
        if not os.path.isdir(self._conda_dir):
            return "Conda dir not created."

        self._progress.setValue(11)
        exe = py_exe(self._conda_dir)
        if not os.path.isfile(exe):
            return "Conda dir does not have Python exe"

        self._progress.setValue(21)
        try:
            ver = subprocess.check_output([exe, "-c", "import sys; print(sys.version)"])
        except Exception as err:
            return "Error getting Python version: " + str(err)

        self._progress.setValue(31)
        if ver.decode() < "3.4":
            return "Expected Python version 3.4 or higher"

        self._progress.setValue(41)
        try:
            ver = subprocess.check_output(
                [exe, "-c", "import conda; print(conda.__version__)"]
            )
        except Exception as err:
            return "Error calling Python exe: " + str(err)

        self._progress.setValue(51)
        if ver.decode() < "3.16":
            return "Expected Conda version 3.16 or higher"

        # Smooth toward 100%
        for i in range(self._progress.value(), 100, 5):
            time.sleep(0.05)
            self._progress.setValue(i)
            QtWidgets.qApp.processEvents()


def is_64bit():
    """Get whether the OS is 64 bit. On WIndows yields what it *really*
    is, not what the process is.
    """
    if False:  # sys.platform.startswith('win'):  ARG, causes problems with subprocess
        if "PROCESSOR_ARCHITEW6432" in os.environ:
            return True
        return os.environ["PROCESSOR_ARCHITECTURE"].endswith("64")
    else:
        return struct.calcsize("P") == 8


def py_exe(dir):
    if sys.platform.startswith("win"):
        return os.path.join(dir, "python.exe")
    else:
        return os.path.join(dir, "bin", "python")


def _chunk_read(response, local_file, chunk_size=1024, initial_size=0, progress=None):
    """Download a file chunk by chunk and show advancement"""
    # Adapted from NISL:
    # https://github.com/nisl/tutorial/blob/master/nisl/datasets.py

    bytes_so_far = initial_size
    # Returns only amount left to download when resuming, not the size of the
    # entire file
    total_size = int(response.headers["Content-Length"].strip())
    total_size += initial_size

    if progress:
        progress.setMaximum(total_size)

    while True:
        QtWidgets.qApp.processEvents()
        chunk = response.read(chunk_size)
        bytes_so_far += len(chunk)
        if not chunk:
            sys.stderr.write("\n")
            break
        # _chunk_write(chunk, local_file, progress)
        progress.setValue(bytes_so_far)
        local_file.write(chunk)


def _fetch_file(url, file_name, progress=None):
    """Load requested file, downloading it if needed or requested"""
    # Adapted from NISL:
    # https://github.com/nisl/tutorial/blob/master/nisl/datasets.py

    temp_file_name = file_name + ".part"
    local_file = None
    initial_size = 0
    try:
        # Checking file size and displaying it alongside the download url
        response = urllib.request.urlopen(url, timeout=5.0)
        # file_size = int(response.headers['Content-Length'].strip())
        # Downloading data (can be extended to resume if need be)
        local_file = open(temp_file_name, "wb")
        _chunk_read(response, local_file, initial_size=initial_size, progress=progress)
        # temp file must be closed prior to the move
        if not local_file.closed:
            local_file.close()
        shutil.move(temp_file_name, file_name)
    except Exception as e:
        raise RuntimeError(
            "Error while fetching file %s.\n" "Dataset fetching aborted (%s)" % (url, e)
        )
    finally:
        if local_file is not None:
            if not local_file.closed:
                local_file.close()


class StreamCatcher(threading.Thread):
    def __init__(self, file, signal):
        self._file = file
        self._signal = signal
        self._lines = []
        self._line = ""
        threading.Thread.__init__(self)
        self.setDaemon(True)  # do not let this thread hold up Python shutdown
        self.start()

    def run(self):
        while True:
            time.sleep(0.0001)
            try:
                part = self._file.read(20)
            except ValueError:  # pragma: no cover
                break
            if not part:
                break
            part = part.decode("utf-8", "ignore")
            # print(part, end='')

            self._line += part.replace("\r", "\n")
            lines = [line for line in self._line.split("\n") if line]
            self._lines.extend(lines[:-1])
            self._line = lines[-1]

            if self._lines:
                self._signal.emit(self._lines[-1])

        self._lines.append(self._line)
        self._signal.emit(self._lines[-1])

    def output(self):
        return "\n".join(self._lines)


if __name__ == "__main__":

    check_for_conda_env()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the (new) BSD License.
# The full license can be found in 'license.txt'.

""" pyzolauncher.py script

This is a script used to startup Pyzo. Added for convenience, and also
for running a test on the source version.

Pyzo can be installed as a package, but it does not have to. You can
start Pyzo in a few different ways:
  * execute this script (pyzolauncher.py)
  * execute the pyzo directory (Python will seek out pyzo/__main__.py)
  * execute the pyzo package ("python -m pyzo")

Only in the latter must Pyzo be installed.

"""

import os
import sys
import subprocess


# faulthandler helps debugging hard crashes, it is included in py3.3
try:
    if sys.executable.lower().endswith("pythonw.exe"):
        raise ImportError("Dont use faulthandler in pythonw.exe")
    import faulthandler

    faulthandler.enable()
except ImportError:
    pass


if "--test" in sys.argv:

    # Prepare log file
    logfilename = os.path.abspath(os.path.join(__file__, "..", "log.txt"))
    with open(logfilename, "wt") as f:
        f.write("")

    # Run Pyzo
    os.environ["PYZO_LOG"] = logfilename
    subprocess.run([sys.executable, "pyzo", "--test"])

else:

    import pyzo

    pyzo.start()



# -*- coding: utf-8 -*-
# Copyright (C) 2016, Almar Klein

"""
Reading and saving Zoof Object Notation files. ZON is like JSON, but a
more Pythonic format. It's just about 500 lines of code.

This format is a spin-off from the SSDF format, it is fully compatible,
except that ZON does not support numpy arrays.
"""

import re
import sys
import time


# From six.py
if sys.version_info[0] >= 3:
    string_types = (str,)
    integer_types = (int,)
else:
    string_types = (basestring,)  # noqa
    integer_types = (int, long)  # noqa
float_types = (float,)


## Dict class

try:  # pragma: no cover
    from collections import OrderedDict as _dict
except ImportError:
    _dict = dict


def isidentifier(s):
    # http://stackoverflow.com/questions/2544972/
    if not isinstance(s, str):
        return False
    return re.match(r"^\w+$", s, re.UNICODE) and re.match(r"^[0-9]", s) is None


class Dict(_dict):
    """A dict in which the items can be get/set as attributes."""

    __reserved_names__ = dir(_dict())  # Also from OrderedDict
    __pure_names__ = dir(dict())

    __slots__ = []

    def __repr__(self):
        identifier_items = []
        nonidentifier_items = []
        for key, val in self.items():
            if isidentifier(key):
                identifier_items.append("%s=%r" % (key, val))
            else:
                nonidentifier_items.append("(%r, %r)" % (key, val))
        if nonidentifier_items:
            return "Dict([%s], %s)" % (
                ", ".join(nonidentifier_items),
                ", ".join(identifier_items),
            )
        else:
            return "Dict(%s)" % (", ".join(identifier_items))

    def __getattribute__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            if key in self:
                return self[key]
            else:
                raise

    def __setattr__(self, key, val):
        if key in Dict.__reserved_names__:
            # Either let OrderedDict do its work, or disallow
            if key not in Dict.__pure_names__:
                return _dict.__setattr__(self, key, val)
            else:
                raise AttributeError(
                    "Reserved name, this key can only "
                    + "be set via ``d[%r] = X``" % key
                )
        else:
            # if isinstance(val, dict): val = Dict(val) -> no, makes a copy!
            self[key] = val

    def __dir__(self):
        names = [k for k in self.keys() if isidentifier(k)]
        return Dict.__reserved_names__ + names


# SSDF compatibility
Struct = Dict
Struct.__is_ssdf_struct__ = True


## Public functions


def isstruct(ob):  # SSDF compatibility
    """isstruct(ob)

    Returns whether the given object is an SSDF struct.
    """
    if hasattr(ob, "__is_ssdf_struct__"):
        return bool(ob.__is_ssdf_struct__)
    else:
        return False


def new():
    """new()

    Create a new Dict object. The same as "Dict()".
    """
    return Dict()


def clear(d):  # SSDF compatibility
    """clear(d)

    Clear all elements of the given Dict object.
    """
    d.clear()


def copy(object):
    """copy(objec)

    Return a deep copy the given object. The object and its children
    should be dict-compatible data types. Note that dicts are converted
    to Dict and tuples to lists.
    """
    if isstruct(object) or isinstance(object, dict):
        newObject = Dict()
        for key in object:
            val = object[key]
            newObject[key] = copy(val)
        return newObject
    elif isinstance(object, (tuple, list)):
        return [copy(ob) for ob in object]
    else:
        return object  # immutable


def count(object, cache=None):
    """count(object):

    Count the number of elements in the given object. An element is
    defined as one of the 6 datatypes supported by ZON (dict,
    tuple/list, string, int, float, None).
    """
    cache = cache or []
    if isstruct(object) or isinstance(object, (dict, list)):
        if id(object) in cache:
            raise RuntimeError("recursion!")
        cache.append(id(object))
    n = 1
    if isstruct(object) or isinstance(object, dict):
        for key in object:
            val = object[key]
            n += count(val, cache)
    elif isinstance(object, (tuple, list)):
        for val in object:
            n += count(val, cache)
    return n


def loads(text):
    """loads(text)

    Load a Dict from the given Unicode) string in ZON syntax.
    """
    if not isinstance(text, string_types):
        raise ValueError("zon.loads() expects a string.")
    reader = ReaderWriter()
    return reader.read(text)


def load(file):
    """load(filename)

    Load a Dict from the given file or filename.
    """
    if isinstance(file, string_types):
        file = open(file, "rb")
    text = file.read().decode("utf-8")
    return loads(text)


def saves(d):
    """saves(d)

    Serialize the given dict to a (Unicode) string.
    """
    if not (isstruct(d) or isinstance(d, dict)):
        raise ValueError("ssdf.saves() expects a dict.")
    writer = ReaderWriter()
    text = writer.save(d)
    return text


def save(file, d):
    """save(file, d)

    Serialize the given dict to the given file or filename.
    """
    text = saves(d)
    if isinstance(file, string_types):
        file = open(file, "wb")
    with file:
        file.write(text.encode("utf-8"))


## The core


class ReaderWriter(object):
    def read(self, text):

        indent = 0
        root = Dict()
        container_stack = [(0, root)]
        new_container = None

        for i, line in enumerate(text.splitlines()):
            linenr = i + 1

            # Strip line
            line2 = line.lstrip()

            # Skip comments and empty lines
            if not line2 or line2[0] == "#":
                continue

            # Find the indentation
            prev_indent = indent
            indent = len(line) - len(line2)
            if indent == prev_indent:
                pass
            elif indent < prev_indent:
                while container_stack[-1][0] > indent:
                    container_stack.pop(-1)
                if container_stack[-1][0] != indent:
                    print("ZON: Ignoring wrong dedentation at %i" % linenr)
            elif indent > prev_indent and new_container is not None:
                container_stack.append((indent, new_container))
                new_container = None
            else:
                print("ZON: Ignoring wrong indentation at %i" % linenr)
                indent = prev_indent

            # Split name and data using a regular expression
            m = re.search("^\w+? *?=", line2)
            if m:
                i = m.end(0)
                name = line2[: i - 1].strip()
                data = line2[i:].lstrip()
            else:
                name = None
                data = line2

            # Get value
            value = self.to_object(data, linenr)

            # Store the value
            _indent, current_container = container_stack[-1]
            if isinstance(current_container, dict):
                if name:
                    current_container[name] = value
                else:
                    print("ZON: unnamed item in dict on line %i" % linenr)
            elif isinstance(current_container, list):
                if name:
                    print("ZON: named item in list on line %i" % linenr)
                else:
                    current_container.append(value)
            else:
                raise RuntimeError("Invalid container %r" % current_container)

            # Prepare for next round
            if isinstance(value, (dict, list)):
                new_container = value

        return root

    def save(self, d):

        pyver = "%i.%i.%i" % sys.version_info[:3]
        ct = time.asctime()
        lines = []
        lines.append("# -*- coding: utf-8 -*-")
        lines.append("# This Zoof Object Notation (ZON) file was")
        lines.append("# created from Python %s on %s.\n" % (pyver, ct))
        lines.append("")
        lines.extend(self.from_dict(d, -2)[1:])

        return "\r\n".join(lines)
        # todo: pop toplevel dict

    def from_object(self, name, value, indent):

        # Get object's data
        if value is None:
            data = "Null"
        elif isinstance(value, integer_types):
            data = self.from_int(value)
        elif isinstance(value, float_types):
            data = self.from_float(value)
        elif isinstance(value, bool):
            data = self.from_int(int(value))
        elif isinstance(value, string_types):
            data = self.from_unicode(value)
        elif isinstance(value, dict):
            data = self.from_dict(value, indent)
        elif isinstance(value, (list, tuple)):
            data = self.from_list(value, indent)
        else:
            # We do not know
            data = "Null"
            tmp = repr(value)
            if len(tmp) > 64:
                tmp = tmp[:64] + "..."
            if name is not None:
                print("ZON: %s is unknown object: %s" % (name, tmp))
            else:
                print("ZON: unknown object: %s" % tmp)

        # Finish line (or first line)
        if isinstance(data, string_types):
            data = [data]
        if name:
            data[0] = "%s%s = %s" % (" " * indent, name, data[0])
        else:
            data[0] = "%s%s" % (" " * indent, data[0])

        return data

    def to_object(self, data, linenr):

        data = data.lstrip()

        # Determine what type of object we're dealing with by reading
        # like a human.
        if not data:
            print("ZON: no value specified at line %i." % linenr)
        elif data[0] in "-.0123456789":
            return self.to_int_or_float(data, linenr)
        elif data[0] == "'":
            return self.to_unicode(data, linenr)
        elif data.startswith("dict:"):
            return self.to_dict(data, linenr)
        elif data.startswith("list:") or data[0] == "[":
            return self.to_list(data, linenr)
        elif data.startswith("Null") or data.startswith("None"):
            return None
        else:
            print("ZON: invalid type on line %i." % linenr)
            return None

    def to_int_or_float(self, data, linenr):
        line = data.partition("#")[0]
        try:
            return int(line)
        except ValueError:
            try:
                return float(line)
            except ValueError:
                print("ZON: could not parse number on line %i." % linenr)
                return None

    def from_int(self, value):
        return repr(int(value)).rstrip("L")

    def from_float(self, value):
        # Use general specifier with a very high precision.
        # Any spurious zeros are automatically removed. The precision
        # should be sufficient such that any numbers saved and loaded
        # back will have the exact same value again.
        # see e.g. http://bugs.python.org/issue1580
        return repr(float(value))  # '%0.17g' % value

    def from_unicode(self, value):
        value = value.replace("\\", "\\\\")
        value = value.replace("\n", "\\n")
        value = value.replace("\r", "\\r")
        value = value.replace("\x0b", "\\x0b").replace("\x0c", "\\x0c")
        value = value.replace("'", "\\'")
        return "'" + value + "'"

    def to_unicode(self, data, linenr):
        # Encode double slashes
        line = data.replace("\\\\", "0x07")  # temp

        # Find string using a regular expression
        m = re.search("'.*?[^\\\\]'|''", line)
        if not m:
            print("ZON: string not ended correctly on line %i." % linenr)
            return None  # return not-a-string
        else:
            line = m.group(0)[1:-1]

        # Decode stuff
        line = line.replace("\\n", "\n")
        line = line.replace("\\r", "\r")
        line = line.replace("\\x0b", "\x0b").replace("\\x0c", "\x0c")
        line = line.replace("\\'", "'")
        line = line.replace("0x07", "\\")
        return line

    def from_dict(self, value, indent):
        lines = ["dict:"]
        # Process children
        for key, val in value.items():
            # Skip all the builtin stuff
            if key.startswith("__"):
                continue
            # Skip methods, or anything else we can call
            if hasattr(val, "__call__"):
                continue  # Note: py3.x does not have function callable
            # Add!
            lines.extend(self.from_object(key, val, indent + 2))
        return lines

    def to_dict(self, data, linenr):
        return Dict()

    def from_list(self, value, indent):
        # Collect subdata and check whether this is a "small list"
        isSmallList = True
        allowedTypes = integer_types + float_types + string_types
        subItems = []
        for element in value:
            if not isinstance(element, allowedTypes):
                isSmallList = False
            subdata = self.from_object(None, element, 0)  # No indent
            subItems.extend(subdata)
        isSmallList = isSmallList and len(subItems) < 256

        # Return data
        if isSmallList:
            return "[%s]" % (", ".join(subItems))
        else:
            data = ["list:"]
            ind = " " * (indent + 2)
            for item in subItems:
                data.append(ind + item)
            return data

    def to_list(self, data, linenr):
        value = []
        if data[0] == "l":  # list:
            return list()
        else:
            i0 = 1
            pieces = []
            inString = False
            escapeThis = False
            line = data
            for i in range(1, len(line)):
                if inString:
                    # Detect how to get out
                    if escapeThis:
                        escapeThis = False
                        continue
                    elif line[i] == "\\":
                        escapeThis = True
                    elif line[i] == "'":
                        inString = False
                else:
                    # Detect going in a string, break, or end
                    if line[i] == "'":
                        inString = True
                    elif line[i] == ",":
                        pieces.append(line[i0:i])
                        i0 = i + 1
                    elif line[i] == "]":
                        piece = line[i0:i]
                        if piece.strip():  # Do not add if empty
                            pieces.append(piece)
                        break
            else:
                print("ZON: short list not closed right on line %i." % linenr)

            # Cut in pieces and process each piece
            value = []
            for piece in pieces:
                v = self.to_object(piece, linenr)
                value.append(v)
            return value



# -*- coding: utf-8 -*-
# Copyright (c) 2016, Almar Klein

"""
This module implements functionality to obtain registered
Python interpreters and to register a Python interpreter in the Windows
registry.

"""

import sys
import os

try:
    import winreg
except ImportError:
    winreg = None

PYTHON_KEY = "SOFTWARE\\Python\\PythonCore"
PYTHON_KEY_WOW64 = "SOFTWARE\\Wow6432Node\\Python\\PythonCore"
INSTALL_KEY = "InstallPath"
PATH_KEY = "PythonPath"


class PythonInReg:
    """Class to represent a specific version of the Python interpreter
    registered (or being registered in the registry).
    This is a helper class for the functions defined in this module; it
    should not be instantiated directly.
    """

    USER_ONE = 1
    USER_ALL = 2

    def __init__(self, user, version, wow64=False):
        self._user = user
        self._key = (wow64 and PYTHON_KEY_WOW64 or PYTHON_KEY) + "\\" + version

    def __repr__(self):
        userstr = [None, "USER_ONE", "USER_ALL"][self._user]
        installPath = self.installPath()
        reg = self._reg()
        if not reg:
            return "<PythonInReg %s at %s (unregistered)>" % (self.version(), userstr)
        elif installPath:
            return '<PythonInReg %s at %s in "%s">' % (
                self.version(),
                userstr,
                installPath,
            )
        else:
            return "<PythonInReg %s at %s>" % (self.version(), userstr)

    def _root(self):
        if self._user == PythonInReg.USER_ONE:
            return winreg.HKEY_CURRENT_USER
        else:
            return winreg.HKEY_LOCAL_MACHINE

    def _reg(self):
        # Get key for this version
        try:
            return winreg.OpenKey(self._root(), self._key, 0, winreg.KEY_READ)
        except Exception:
            return None

    def create(self):
        """Create key. If already exists, does nothing."""

        # Get key for this version
        reg = self._reg()

        if reg:
            winreg.CloseKey(reg)
            # print('Unable to create Python version %s: already exists.' % self.version())

        else:
            # Try to create
            try:
                reg = winreg.CreateKey(self._root(), self._key)
                winreg.CloseKey(reg)
            except Exception:
                raise RuntimeError(
                    "Unable to create python version %s." % self.version()
                )
            print("Created %s." % str(self))

    def delete(self):

        # Get key for this version
        reg = self._reg()
        if not reg:
            print("Unable to delete Python version %s: does not exist.")

        # Delete attributes
        try:
            winreg.DeleteKey(reg, INSTALL_KEY)
        except Exception:
            pass
        try:
            winreg.DeleteKey(reg, PATH_KEY)
        except Exception:
            pass

        # Delete main key for this version, or show warning
        try:
            winreg.DeleteKey(self._root(), self._key)
        except Exception:
            print("Could not delete %s." % str(self))
            return
        print("Deleted %s." % str(self))

    def setInstallPath(self, installPath):
        # Get key for this version
        reg = self._reg()
        if not reg:
            raise RuntimeError(
                "Could not set installPath for version %s: version does not exist."
                % self.version()
            )

        # Set value or raise error
        try:
            winreg.SetValue(reg, INSTALL_KEY, winreg.REG_SZ, installPath)
            winreg.CloseKey(reg)
        except Exception:
            winreg.CloseKey(reg)
            raise RuntimeError("Could not set installPath for %s." % str(self))

    def installPath(self):
        # Get key for this version
        reg = self._reg()
        if not reg:
            return None

        # Get value or return None
        try:
            installPath = winreg.QueryValue(reg, INSTALL_KEY)
            winreg.CloseKey(reg)
            return installPath
        except Exception:
            winreg.CloseKey(reg)
            return None

    def setPythonPath(self, pythonPath):
        # Get key for this version
        reg = self._reg()
        if not reg:
            raise RuntimeError(
                "Could not set pythonPath for version %s: version does not exist."
                % self.version()
            )

        # Set value or raise error
        try:
            winreg.SetValue(reg, PATH_KEY, winreg.REG_SZ, pythonPath)
            winreg.CloseKey(reg)
        except Exception:
            winreg.CloseKey(reg)
            raise RuntimeError("Could not set pythonPath for %s." % str(self))

    def pythonPath(self):
        # Get key for this version
        reg = self._reg()
        if not reg:
            return None

        # Get value or return None
        try:
            pythonPath = winreg.QueryValue(reg, PATH_KEY)
            winreg.CloseKey(reg)
            return pythonPath
        except Exception:
            winreg.CloseKey(reg)
            return None

    def version(self):
        """Get the Python version."""
        return self._key[-3:]


def get_interpreters_in_reg():
    """get_interpreters_in_reg()
    Get a list of PythonInReg instances: one for each interpreter
    in the registry. This function checks both LOCAL_MACHINE and CURRENT_USER.
    """
    versions = []
    for user in [1, 2]:
        for wow64 in [False, True]:
            versions.extend(_get_interpreter_in_reg(user, wow64))
    return versions


def _get_interpreter_in_reg(user, wow64=False):

    # Get base key
    if user == PythonInReg.USER_ONE:
        HKEY = winreg.HKEY_CURRENT_USER
    else:
        HKEY = winreg.HKEY_LOCAL_MACHINE

    # Get Python key
    if wow64:
        PYKEY = PYTHON_KEY_WOW64
    else:
        PYKEY = PYTHON_KEY

    # Try to open Python key
    try:
        reg = winreg.OpenKey(HKEY, PYKEY, 0, winreg.KEY_READ)
    except Exception:
        return []

    # Get info about subkeys
    nsub, nval, modified = winreg.QueryInfoKey(reg)

    # Query all
    versions = []
    for i in range(nsub):
        # Get name and subkey
        version = winreg.EnumKey(reg, i)
        versions.append(PythonInReg(user, version, wow64))

    # Done
    winreg.CloseKey(reg)
    return versions


def register_interpreter(version=None, installPath=None, user=None, wow64=False):
    """register_interpreter(version=None, installPath=None, user=None, wow64=False)
    Register a certain Python version. If version and installPath
    are not given, the current Python process is registered.
    if user is not given, tries LOCAL_MACHINE first but uses CURRENT_USER
    if that fails.
    """
    if version is None:
        version = sys.version[:3]
    if installPath is None:
        installPath = sys.prefix

    # Get existing versions
    existingVersions = get_interpreters_in_reg()

    # Determine what users to try
    if user is None:
        users = [2, 1]
    else:
        users = [user]

    success = False

    for user in users:

        # Create new PythonInReg instance
        v = PythonInReg(user, version, wow64)

        # Check if already exists
        ok = True
        for ev in existingVersions:
            if ev._key != v._key or ev._user != v._user:
                continue  # Different key; no problem
            if (not ev.installPath()) or (not os.path.isdir(ev.installPath())):
                continue  # Key the same, but existing entry is invalid
            if ev.installPath() == installPath:
                # Exactly the same, no action required, return now!
                return ev
            # Ok, there's a problem
            ok = False
            print(
                'Warning: version %s is already installed in "%s".'
                % (version, ev.installPath())
            )
        if not ok:
            continue

        # Try to create the key
        try:
            v.create()
            v.setInstallPath(installPath)
            success = True
            break
        except RuntimeError:
            continue

    if success:
        return v
    else:
        raise RuntimeError(
            "Could not register Python version %s at %s." % (version, installPath)
        )


if __name__ == "__main__":

    for v in get_interpreters_in_reg():
        print(v)



""" _utest.py

Unit tests for yoton.

"""

import sys, os, time
import unittest

# Go up one directory and then import the yoton package
os.chdir("..")
os.chdir("..")
sys.path.insert(0, ".")

# Import yoton from there
import yoton
from yoton.misc import long


class Tester(unittest.TestCase):
    """Perform a set of tests using three contexts and
    each of the six channels at each context.

    """

    def setUp(self):

        # Create contexts
        self._context1 = c1 = yoton.Context()
        self._context2 = c2 = yoton.Context()
        self._context3 = c3 = yoton.Context()

        # Create pub sub channels
        self._channel_pub1 = yoton.PubChannel(c1, "foo")
        self._channel_pub2 = yoton.PubChannel(c2, "foo")
        self._channel_pub3 = yoton.PubChannel(c3, "foo")
        #
        self._channel_sub1 = yoton.SubChannel(c1, "foo")
        self._channel_sub2 = yoton.SubChannel(c2, "foo")
        self._channel_sub3 = yoton.SubChannel(c3, "foo")

        # Create req rep channels
        self._channel_req1 = yoton.ReqChannel(c1, "bar")
        self._channel_req2 = yoton.ReqChannel(c2, "bar")
        self._channel_req3 = yoton.ReqChannel(c3, "bar")
        #
        self._channel_rep1 = yoton.RepChannel(c1, "bar")
        self._channel_rep2 = yoton.RepChannel(c2, "bar")
        self._channel_rep3 = yoton.RepChannel(c3, "bar")

        # Create state channels
        self._channel_state1 = yoton.StateChannel(c1, "spam")
        self._channel_state2 = yoton.StateChannel(c2, "spam")
        self._channel_state3 = yoton.StateChannel(c3, "spam")

    def tearDown(self):
        self._context1.close()
        self._context2.close()
        self._context3.close()

    def test_connecting(self):

        # Connect
        self._context1.bind("localhost:test1")
        self._context2.connect("localhost:test1")
        time.sleep(0.1)  # Give time for binding to finish

        # Test if connected
        self.assertEqual(self._context1.connection_count, 1)
        self.assertEqual(self._context2.connection_count, 1)
        self.assertEqual(self._context3.connection_count, 0)

        # Connect more
        self._context1.bind("localhost:test2")
        self._context3.connect("localhost:test2")
        time.sleep(0.1)  # Give time for binding to finish

        # Test if connected
        self.assertEqual(self._context1.connection_count, 2)
        self.assertEqual(self._context2.connection_count, 1)
        self.assertEqual(self._context3.connection_count, 1)

    def test_closing_channel(self):

        # Connect
        self._context1.bind("localhost:test")
        self._context2.connect("localhost:test")

        # Send data
        self._channel_pub1.send("hello")

        # Receive data
        self.assertEqual(self._channel_sub2.recv(), "hello")
        self.assertEqual(self._channel_sub2.recv(False), "")

        # Close channel
        self._channel_sub2.close()
        self.assertEqual(self._channel_sub2.recv(), "")
        self.assertTrue(self._channel_sub2.closed)

    def test_pub_sub(self):

        # also test hopping

        # Connect
        self._context1.bind("localhost:test1")
        self._context2.connect("localhost:test1")
        self._context2.bind("localhost:test2")
        self._context3.connect("localhost:test2")

        # Send messages
        self._channel_pub1.send("msg1")
        time.sleep(0.1)
        self._channel_pub2.send("msg2")
        time.sleep(0.1)

        # Receive msssages
        self.assertEqual(self._channel_sub2.recv(), "msg1")
        self.assertEqual(self._channel_sub3.recv(), "msg1")
        self.assertEqual(self._channel_sub3.recv(), "msg2")

    def test_pub_sub_select_channel(self):

        # Connect
        self._context1.bind("localhost:test1")
        self._context2.connect("localhost:test1")

        # Create channels
        pub1 = yoton.PubChannel(self._context1, "foo1")
        pub2 = yoton.PubChannel(self._context1, "foo2")
        #
        sub1 = yoton.SubChannel(self._context2, "foo1")
        sub2 = yoton.SubChannel(self._context2, "foo2")

        # Send a bunch of messages
        ii = [0, 1, 0, 0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1]
        for i in range(len(ii)):
            pub = [pub1, pub2][ii[i]]
            pub.send(str(i))

        time.sleep(0.1)

        # Receive in right order
        count = 0
        while True:
            sub = yoton.select_sub_channel(sub1, sub2)
            if sub:
                i = int(sub.recv())
                self.assertEqual(i, count)
                count += 1
            else:
                break

        # Test count
        self.assertEqual(count, len(ii))

    def test_pubstate_substate(self):

        # also test hopping

        # Set status before connecting
        self._channel_state1.send("status1")

        # Connect
        self._context1.bind("localhost:test1")
        self._context2.connect("localhost:test1")
        self._context2.bind("localhost:test2")
        self._context3.connect("localhost:test2")
        time.sleep(0.1)

        # Receive msssages
        self.assertEqual(self._channel_state2.recv(), "status1")
        self.assertEqual(self._channel_state2.recv(), "status1")
        self.assertEqual(self._channel_state3.recv(), "status1")
        self.assertEqual(self._channel_state3.recv(), "status1")

        # Update status
        self._channel_state1.send("status2")
        time.sleep(0.1)

        # Receive msssages (recv multiple times should return last status)
        self.assertEqual(self._channel_state2.recv(), "status2")
        self.assertEqual(self._channel_state2.recv(), "status2")
        self.assertEqual(self._channel_state3.recv(), "status2")
        self.assertEqual(self._channel_state3.recv(), "status2")

    def test_req_rep1(self):

        # Connect
        self._context1.bind("localhost:test1")
        self._context2.connect("localhost:test1")
        self._context2.bind("localhost:test2")
        self._context3.connect("localhost:test2")

        # Turn on one replier
        self._channel_rep3.set_mode("thread")

        # Do requests
        foo, id1 = self._channel_req1.echo("foo").result(1)
        bar, id2 = self._channel_req1.echo("bar").result(1)

        # Check reply
        self.assertEqual(foo, "foo")
        self.assertEqual(bar, "bar")
        self.assertEqual(id1, hex(self._context3.id))
        self.assertEqual(id2, hex(self._context3.id))

    def test_req_rep2(self):

        # Test multiple requesters and multiple repliers

        # Connect
        if True:
            self._context1.bind("localhost:test1")
            self._context3.connect("localhost:test1")
            #
            self._context1.bind("localhost:test2")
            self._context2.connect("localhost:test2")
        else:
            freeNode = yoton.Context()
            #
            self._context1.bind("localhost:test1")
            freeNode.connect("localhost:test1")
            freeNode.bind("localhost:test1f")
            self._context3.connect("localhost:test1f")
            #
            self._context1.bind("localhost:test2")
            self._context2.connect("localhost:test2")
        time.sleep(0.1)

        # Turn on 2 requesters and 3 repliers
        self._channel_rep1.set_mode("thread")  # threaded because simulating
        self._channel_rep2.set_mode("thread")  # different processes
        self._channel_rep3.set_mode("thread")

        # Define and register reply handler
        def reply_handler(future):
            reply = future.result()
            reqnr = future.reqnr

            echo, id = reply
            if echo.lower() == "stop":
                yoton.stop_event_loop()
            else:
                self.assertEqual(echo[:3], "msg")
                contextnr = {
                    self._context1.id: 1,
                    self._context2.id: 2,
                    self._context3.id: 3,
                }[long(id, 16)]
                print(
                    "request %s from %i handled by context %i."
                    % (echo, reqnr, contextnr)
                )

        # Get echo functions
        echoFun1 = self._channel_req1.echo
        echoFun2 = self._channel_req2.echo

        # Send requests on req 1
        sleepTimes = [1, 0.1, 1, 0.6, 0.6, 0.6, 0.6, 0.6]
        for i in range(len(sleepTimes)):
            f = echoFun1("msg%i" % i, sleepTimes[i])
            f.add_done_callback(reply_handler)
            f.reqnr = 1

        # Send requests on req 2
        sleepTimes = [0.4, 0.4, 0.4, 0.4, 0.4]
        for i in range(len(sleepTimes)):
            f = echoFun2("msg%i" % i, sleepTimes[i])
            f.add_done_callback(reply_handler)
            f.reqnr = 2

        # Stop
        f = echoFun1("stop")
        f.add_done_callback(reply_handler)
        f.reqnr = 1

        # Enter event loop
        yoton.start_event_loop()


if __name__ == "__main__":

    suite = unittest.TestLoader().loadTestsFromTestCase(Tester)
    unittest.TextTestRunner(verbosity=2).run(suite)
    print("")



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the codeeditor development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

"""
The base code editor class.


"""

"""
WRITING EXTENSIONS FOR THE CODE EDITOR

The Code Editor extension mechanism works solely based on inheritance.
Extensions can override event handlers (e.g. paintEvent, keyPressEvent). Their
default behaviour should be to call their super() event handler. This way,
events propagate through the extensions following Python's method resolution
order (http://www.python.org/download/releases/2.3/mro/).

A 'fancy' code editor with extensions is created like:

class FancyEditor( Extension1, Extension2, ... CodeEditorBase):
    pass

The order of the extensions does usually matter! If multiple Extensions process
the same key press, the first one has the first chance to consume it.

OVERRIDING __init__

An extensions' __init__ method (if required) should look like this:
class Extension:
    def __init__(self, *args, extensionParam1 = 1, extensionParam2 = 3, **kwds):
        super().__init__(*args, **kwds)
        some_extension_init_stuff()
        
Note the following points:
 - All parameters have default values
 - The use of *args passes all non-named arguments to its super(), which
   will therefore end up at the QPlainTextEdit constructor. As a consequence,
   the parameters of the exentsion can only be specified as named arguments
 - The use of **kwds ensures that parametes that are not defined by this
   extension, are passed to the next extension(s) in line.
 - The call to super().__init__ is the first thing to do, this ensures that at
   least the CodeEditorBase and QPlainTextEdit, of which the CodeEditorBase is
   derived, are initialized when the initialization of the extension is done

OVERRIDING keyPressEvent

When overriding keyPressEvent, the extension has several options when an event
arrives:
 - Ignore the event
     In this case, call super().keyPressEvent(event) for other extensions or the
     CodeEditorBase to process the event
 - Consume the event
     In order to prevent other next extensions or the CodeEditorBase to react
     on the event, return without calling the super().keyPressEvent
 - Do something based on the event, and do not let the event propagate
     In this case, do whatever action is defined by the extension, and do not
     call the super().keyPressEvent
 - Do something based on the event, and let the event propagate
     In this case, do whatever action is defined by the extension, and do call
     the super().keyEvent

In any case, the keyPressEvent should not return a value (i.e., return None).
Furthermore, an extension may also want to perform some action *after* the
event has been processed by the next extensions and the CodeEditorBase. In this
case, perform that action after calling super().keyPressEvent

OVERRIDING paintEvent

Then overriding the paintEvent, the extension may want to paint either behind or
in front of the CodeEditorBase text. In order to paint behind the text, first
perform the painting, and then call super().paintEvent. In order to paint in
front of the text, first call super().paintEvent, then perform the painting.

As a result, the total paint order is as follows for the example of the
FancyEditor defined above:
- First the extensions that draw behind the text (i.e. paint before calling
  super().paintEvent, in the order Extension1, Extension2, ...
- then the CodeEditorBase, with the text
- then the extensions that draw in front of the text (i.e. call
  super().paintEvent before painting), in the order ..., Extension2, Extension1
  
OVERRIDING OTHER EVENT HANDLERS

When overriding other event handlers, be sure to call the super()'s event
handler; either before or after your own actions, as appropriate

OTHER ISSUES

In order to avoid namespace clashes among the extensions, take the following
into account:
 - Private members should start with __ to make ensure no clashes will occur
 - Public members / methods should have names that clearly indicate which
   extension they belong to (e.g. not cancel but autocompleteCancel)
 - Arguments of the __init__ method should also have clearly destictive names

"""

from .qt import QtGui, QtCore, QtWidgets

Qt = QtCore.Qt

from .misc import DEFAULT_OPTION_NAME, DEFAULT_OPTION_NONE, ce_option
from .misc import callLater, ustr
from .manager import Manager
from .highlighter import Highlighter
from .style import StyleElementDescription


class CodeEditorBase(QtWidgets.QPlainTextEdit):
    """The base code editor class. Implements some basic features required
    by the extensions.

    """

    # Style element for default text and editor background
    _styleElements = [
        (
            "Editor.text",
            "The style of the default text. "
            + "One can set the background color here.",
            "fore:#000,back:#fff",
        )
    ]

    # Signal emitted after style has changed
    styleChanged = QtCore.Signal()

    # Signal emitted after font (or font size) has changed
    fontChanged = QtCore.Signal()

    # Signal to indicate a change in breakpoints. Only emitted if the
    # appropriate extension is in use
    breakPointsChanged = QtCore.Signal(object)

    def __init__(self, *args, **kwds):
        super(CodeEditorBase, self).__init__(*args)

        # Set font (always monospace)
        self.__zoom = 0
        self.setFont()

        # Create highlighter class
        self.__highlighter = Highlighter(self, self.document())

        # Set some document options
        option = self.document().defaultTextOption()
        option.setFlags(
            option.flags()
            | option.IncludeTrailingSpaces
            | option.AddSpaceForLineAndParagraphSeparators
        )
        self.document().setDefaultTextOption(option)

        # When the cursor position changes, invoke an update, so that
        # the hihghlighting etc will work
        self.cursorPositionChanged.connect(self.viewport().update)

        # Init styles to default values
        self.__style = {}
        for element in self.getStyleElementDescriptions():
            self.__style[element.key] = element.defaultFormat

        # Connext style update
        self.styleChanged.connect(self.__afterSetStyle)
        self.__styleChangedPending = False

        # Init margins
        self._leftmargins = []

        # Init options now.
        # NOTE TO PEOPLE DEVELOPING EXTENSIONS:
        # If an extension has an __init__ in which it first calls the
        # super().__init__, this __initOptions() function will be called,
        # while the extension's init is not yet finished.
        self.__initOptions(kwds)

        # Define colors from Solarized theme
        base03 = "#002b36"
        base02 = "#073642"
        base01 = "#586e75"
        base00 = "#657b83"
        base0 = "#839496"
        base1 = "#93a1a1"
        base2 = "#eee8d5"
        base3 = "#fdf6e3"
        yellow = "#b58900"
        orange = "#cb4b16"
        red = "#dc322f"  # noqa
        magenta = "#d33682"
        violet = "#6c71c4"
        blue = "#268bd2"
        cyan = "#2aa198"
        green = "#859900"  # noqa

        if True:  # Light vs dark
            # back1, back2, back3 = base3, base2, base1 # real solarised
            back1, back2, back3 = "#fff", base2, base1  # crispier
            fore1, fore2, fore3, fore4 = base00, base01, base02, base03
        else:
            back1, back2, back3 = base03, base02, base01
            fore1, fore2, fore3, fore4 = base0, base1, base2, base3  # noqa

        # Define style using "Solarized" colors
        S = {}
        S["Editor.text"] = "back:%s, fore:%s" % (back1, fore1)
        S["Syntax.identifier"] = "fore:%s, bold:no, italic:no, underline:no" % fore1
        S["Syntax.nonidentifier"] = "fore:%s, bold:no, italic:no, underline:no" % fore2
        S["Syntax.keyword"] = "fore:%s, bold:yes, italic:no, underline:no" % fore2

        S["Syntax.builtins"] = "fore:%s, bold:no, italic:no, underline:no" % fore1
        S["Syntax.instance"] = "fore:%s, bold:no, italic:no, underline:no" % fore1

        S["Syntax.functionname"] = "fore:%s, bold:yes, italic:no, underline:no" % fore3
        S["Syntax.classname"] = "fore:%s, bold:yes, italic:no, underline:no" % orange

        S["Syntax.string"] = "fore:%s, bold:no, italic:no, underline:no" % violet
        S["Syntax.unterminatedstring"] = (
            "fore:%s, bold:no, italic:no, underline:dotted" % violet
        )
        S["Syntax.python.multilinestring"] = (
            "fore:%s, bold:no, italic:no, underline:no" % blue
        )

        S["Syntax.number"] = "fore:%s, bold:no, italic:no, underline:no" % cyan
        S["Syntax.comment"] = "fore:%s, bold:no, italic:no, underline:no" % yellow
        S["Syntax.todocomment"] = "fore:%s, bold:no, italic:yes, underline:no" % magenta
        S["Syntax.python.cellcomment"] = (
            "fore:%s, bold:yes, italic:no, underline:full" % yellow
        )

        S["Editor.Long line indicator"] = "linestyle:solid, fore:%s" % back2
        S["Editor.Highlight current line"] = "back:%s" % back2
        S["Editor.Indentation guides"] = "linestyle:solid, fore:%s" % back2
        S["Editor.Line numbers"] = "back:%s, fore:%s" % (back2, back3)

        # Apply a good default style
        self.setStyle(S)

    # see https://bugreports.qt.io/browse/QTBUG-57552?focusedCommentId=469402&page=com.atlassian.jira.plugin.system.issuetabpanels%3Acomment-tabpanel#comment-469402
    # and https://code.qt.io/cgit/qt/qtbase.git/tree/src/gui/text/qtextdocument.cpp#n1183
    # and https://doc.qt.io/qt-5/qchar.html
    _plainTextTrans = str.maketrans(
        {"\u2029": "\n", "\u2028": "\n", "\ufdd0": "\n", "\ufdd1": "\n"}
    )

    def toPlainText(self):
        cursor = self.textCursor()
        cursor.select(QtGui.QTextCursor.Document)
        return cursor.selectedText().translate(self._plainTextTrans)

    def _setHighlighter(self, highlighterClass):
        self.__highlighter = highlighterClass(self, self.document())

    ## Options

    def __getOptionSetters(self):
        """Get a dict that maps (lowercase) option names to the setter
        methods.
        """

        # Get all names that can be options
        allNames = set(dir(self))
        nativeNames = set(dir(QtWidgets.QPlainTextEdit))
        names = allNames.difference(nativeNames)

        # Init dict of setter members
        setters = {}

        for name in names:
            # Get name without set
            if name.lower().startswith("set"):
                name = name[3:]
            # Get setter and getter name
            name_set = "set" + name[0].upper() + name[1:]
            name_get = name[0].lower() + name[1:]
            # Check if both present
            if not (name_set in names and name_get in names):
                continue
            # Get members
            member_set = getattr(self, name_set)
            member_get = getattr(self, name_get)
            # Check if option decorator was used and get default value
            for member in [member_set, member_get]:
                if hasattr(member, DEFAULT_OPTION_NAME):
                    defaultValue = member.__dict__[DEFAULT_OPTION_NAME]
                    break
            else:
                continue
            # Set default on both
            member_set.__dict__[DEFAULT_OPTION_NAME] = defaultValue
            member_get.__dict__[DEFAULT_OPTION_NAME] = defaultValue
            # Add to list
            setters[name.lower()] = member_set

        # Done
        return setters

    def __setOptions(self, setters, options):
        """Sets the options, given the list-of-tuples methods and an
        options dict.
        """

        # List of invalid keys
        invalidKeys = []

        # Set options
        for key1 in options:
            key2 = key1.lower()
            # Allow using the setter name
            if key2.startswith("set"):
                key2 = key2[3:]
            # Check if exists. If so, call!
            if key2 in setters:
                fun = setters[key2]
                val = options[key1]
                fun(val)
            else:
                invalidKeys.append(key1)

        # Check if invalid keys were given
        if invalidKeys:
            print("Warning, invalid options given: " + ", ".join(invalidKeys))

    def __initOptions(self, options=None):
        """Init the options with their default values.
        Also applies the docstrings of one to the other.
        """

        # Make options an empty dict if not given
        if not options:
            options = {}

        # Get setters
        setters = self.__getOptionSetters()

        # Set default value
        for member_set in setters.values():
            defaultVal = member_set.__dict__[DEFAULT_OPTION_NAME]
            if defaultVal != DEFAULT_OPTION_NONE:
                try:
                    member_set(defaultVal)
                except Exception:
                    print("Error initing option ", member_set.__name__)

        # Also set using given opions?
        if options:
            self.__setOptions(setters, options)

    def setOptions(self, options=None, **kwargs):
        """setOptions(options=None, **kwargs)

        Set the code editor options (e.g. highlightCurrentLine) using
        a dict-like object, or using keyword arguments (options given
        in the latter overrule opions in the first).

        The keys in the dict are case insensitive and one can use the
        option's setter or getter name.

        """

        # Process options
        if options:
            D = {}
            for key in options:
                D[key] = options[key]
            D.update(kwargs)
        else:
            D = kwargs

        # Get setters
        setters = self.__getOptionSetters()

        # Go
        self.__setOptions(setters, D)

    ## Font

    def setFont(self, font=None):
        """setFont(font=None)

        Set the font for the editor. Should be a monospace font. If not,
        Qt will select the best matching monospace font.

        """

        defaultFont = Manager.defaultFont()

        # Get font object
        if font is None:
            font = defaultFont
        elif isinstance(font, QtGui.QFont):
            pass
        elif isinstance(font, str):
            font = QtGui.QFont(font)
        else:
            raise ValueError("setFont accepts None, QFont or string.")

        # Hint Qt that it should be monospace
        font.setStyleHint(font.TypeWriter, font.PreferDefault)

        # Get family, fall back to default if qt could not produce monospace
        fontInfo = QtGui.QFontInfo(font)
        if fontInfo.fixedPitch():
            family = fontInfo.family()
        else:
            family = defaultFont.family()

        # Get size: default size + zoom
        size = defaultFont.pointSize() + self.__zoom

        # Create font instance
        font = QtGui.QFont(family, size)

        # Set, emit and return
        QtWidgets.QPlainTextEdit.setFont(self, font)
        self.fontChanged.emit()
        return font

    def setZoom(self, zoom):
        """setZoom(zoom)

        Set the zooming of the document. The font size is always the default
        font size + the zoom factor.

        The final zoom is returned, this may not be the same as the given
        zoom factor if the given factor is too small.

        """
        # Set zoom (limit such that final pointSize >= 1)
        size = Manager.defaultFont().pointSize()
        self.__zoom = int(max(1 - size, zoom))
        # Set font
        self.setFont(self.fontInfo().family())
        # Return zoom
        return self.__zoom

    ## Syntax / styling

    @classmethod
    def getStyleElementDescriptions(cls):
        """getStyleElementDescriptions()

        This classmethod returns a list of the StyleElementDescription
        instances used by this class. This includes the descriptions for
        the syntax highlighting of all parsers.

        """

        # Collect members by walking the class bases
        elements = []

        def collectElements(cls, iter=1):
            # Valid class?
            if cls is object or cls is QtWidgets.QPlainTextEdit:
                return
            # Check members
            if hasattr(cls, "_styleElements"):
                for element in cls._styleElements:
                    elements.append(element)
            # Recurse
            for c in cls.__bases__:
                collectElements(c, iter + 1)

        collectElements(cls)

        # Make style element descriptions
        # (Use a dict to ensure there are no duplicate keys)
        elements2 = {}
        for element in elements:
            # Check
            if isinstance(element, StyleElementDescription):
                pass
            elif isinstance(element, tuple):
                element = StyleElementDescription(*element)
            else:
                print("Warning: invalid element: " + repr(element))
            # Store using the name as a key to prevent duplicates
            elements2[element.key] = element

        # Done
        return list(elements2.values())

    def getStyleElementFormat(self, name):
        """getStyleElementFormat(name)

        Get the style format for the style element corresponding with
        the given name. The name is case insensitive and invariant to
        the use of spaces.

        """
        key = name.replace(" ", "").lower()
        try:
            return self.__style[key]
        except KeyError:
            raise KeyError('Not a known style element name: "%s".' % name)

    def setStyle(self, style=None, **kwargs):
        """setStyle(style=None, **kwargs)

        Updates the formatting per style element.

        The style consists of a dictionary that maps style names to
        style formats. The style names are case insensitive and invariant
        to the use of spaces.

        For convenience, keyword arguments may also be used. In this case,
        underscores are interpreted as dots.

        This function can also be called without arguments to force the
        editor to restyle (and rehighlight) itself.

        Use getStyleElementDescriptions() to get information about the
        available styles and their default values.

        Examples
        --------
        # To make the classname in underline, but keep the color and boldness:
        setStyle(syntax_classname='underline')
        # To set all values for function names:
        setStyle(syntax_functionname='#883,bold:no,italic:no')
        # To set line number and indent guides colors
        setStyle({  'editor.LineNumbers':'fore:#000,back:#777',
                    'editor.indentationGuides':'#f88' })

        """

        # Combine user input
        D = {}
        if style:
            for key in style:
                D[key] = style[key]
        if True:
            for key in kwargs:
                key2 = key.replace("_", ".")
                D[key2] = kwargs[key]

        # List of given invalid style element names
        invalidKeys = []

        # Set style elements
        for key in D:
            normKey = key.replace(" ", "").lower()
            if normKey in self.__style:
                # self.__style[normKey] = StyleFormat(D[key])
                self.__style[normKey].update(D[key])
            else:
                invalidKeys.append(key)

        # Give warning for invalid keys
        if invalidKeys:
            print("Warning, invalid style names given: " + ",".join(invalidKeys))

        # Notify that style changed, adopt a lazy approach to make loading
        # quicker.
        if self.isVisible():
            callLater(self.styleChanged.emit)
            self.__styleChangedPending = False
        else:
            self.__styleChangedPending = True

    def showEvent(self, event):
        super(CodeEditorBase, self).showEvent(event)
        # Does the style need updating?
        if self.__styleChangedPending:
            callLater(self.styleChanged.emit)
            self.__styleChangedPending = False

    def __afterSetStyle(self):
        """_afterSetStyle()

        Method to call after the style has been set.

        """

        # Set text style using editor style sheet
        format = self.getStyleElementFormat("editor.text")
        ss = "QPlainTextEdit{ color:%s; background-color:%s; }" % (
            format["fore"],
            format["back"],
        )
        self.setStyleSheet(ss)

        # Make sure the style is applied
        self.viewport().update()

        # Re-highlight
        callLater(self.__highlighter.rehighlight)

    ## Some basic options

    @ce_option(4)
    def indentWidth(self):
        """Get the width of a tab character, and also the amount of spaces
        to use for indentation when indentUsingSpaces() is True.
        """
        return self.__indentWidth

    def setIndentWidth(self, value):
        value = int(value)
        if value <= 0:
            raise ValueError("indentWidth must be >0")
        self.__indentWidth = value
        self.setTabStopWidth(self.fontMetrics().width("i" * self.__indentWidth))

    @ce_option(False)
    def indentUsingSpaces(self):
        """Get whether to use spaces (if True) or tabs (if False) to indent
        when the tab key is pressed
        """
        return self.__indentUsingSpaces

    def setIndentUsingSpaces(self, value):
        self.__indentUsingSpaces = bool(value)
        self.__highlighter.rehighlight()

    ## Misc

    def gotoLine(self, lineNumber):
        """gotoLine(lineNumber)

        Move the cursor to the block given by the line number
        (first line is number 1) and show that line.

        """
        return self.gotoBlock(lineNumber - 1)

    def gotoBlock(self, blockNumber):
        """gotoBlock(blockNumber)

        Move the cursor to the block given by the block number
        (first block is number 0) and show that line.

        """
        # Two implementatios. I know that the latter works, so lets
        # just use that.

        cursor = self.textCursor()
        # block = self.document().findBlockByNumber( blockNumber )
        # cursor.setPosition(block.position())
        cursor.movePosition(cursor.Start)  # move to begin of the document
        cursor.movePosition(cursor.NextBlock, n=blockNumber)  # n blocks down

        try:
            self.setTextCursor(cursor)
        except Exception:
            pass  # File is smaller then the caller thought

        # TODO make this user configurable (setting relativeMargin to anything above
        # 0.5 will cause cursor to center on each move)
        relativeMargin = 0.2  # 20% margin on both sides of the window
        margin = self.height() * relativeMargin
        cursorRect = self.cursorRect(cursor)
        if cursorRect.top() < margin or cursorRect.bottom() + margin > self.height():
            self.centerCursor()

    def doForSelectedBlocks(self, function):
        """doForSelectedBlocks(function)

        Call the given function(cursor) for all blocks in the current selection
        A block is considered to be in the current selection if a part of it is in
        the current selection

        The supplied cursor will be located at the beginning of each block. This
        cursor may be modified by the function as required

        """

        # Note: a 'TextCursor' does not represent the actual on-screen cursor, so
        # movements do not move the on-screen cursor

        # Note 2: when the text is changed, the cursor and selection start/end
        # positions of all cursors are updated accordingly, so the screenCursor
        # stays in place even if characters are inserted at the editCursor

        screenCursor = self.textCursor()  # For maintaining which region is selected
        editCursor = self.textCursor()  # For inserting the comment marks

        # Use beginEditBlock / endEditBlock to make this one undo/redo operation
        editCursor.beginEditBlock()

        try:
            editCursor.setPosition(screenCursor.selectionStart())
            editCursor.movePosition(editCursor.StartOfBlock)
            # < :if selection end is at beginning of the block, don't include that
            # one, except when the selectionStart is same as selectionEnd
            while (
                editCursor.position() < screenCursor.selectionEnd()
                or editCursor.position() <= screenCursor.selectionStart()
            ):
                # Create a copy of the editCursor and call the user-supplied function
                editCursorCopy = QtGui.QTextCursor(editCursor)
                function(editCursorCopy)

                # Move to the next block
                if not editCursor.block().next().isValid():
                    break  # We reached the end of the document
                editCursor.movePosition(editCursor.NextBlock)
        finally:
            editCursor.endEditBlock()

    def doForVisibleBlocks(self, function):
        """doForVisibleBlocks(function)

        Call the given function(cursor) for all blocks that are currently
        visible. This is used by several appearence extensions that
        paint per block.

        The supplied cursor will be located at the beginning of each block. This
        cursor may be modified by the function as required

        """

        # Start cursor at top line.
        cursor = self.cursorForPosition(QtCore.QPoint(0, 0))
        cursor.movePosition(cursor.StartOfBlock)

        if not self.isVisible():
            return

        while True:
            # Call the function with a copy of the cursor
            function(QtGui.QTextCursor(cursor))

            # Go to the next block (or not if we are done)
            y = self.cursorRect(cursor).bottom()
            if y > self.height():
                break  # Reached end of the repaint area
            if not cursor.block().next().isValid():
                break  # Reached end of the text
            cursor.movePosition(cursor.NextBlock)

    def indentBlock(self, cursor, amount=1):
        """indentBlock(cursor, amount=1)

        Indent the block given by cursor.

        The cursor specified is used to do the indentation; it is positioned
        at the beginning of the first non-whitespace position after completion
        May be overridden to customize indentation.

        """
        text = ustr(cursor.block().text())
        leadingWhitespace = text[: len(text) - len(text.lstrip())]

        # Select the leading whitespace
        cursor.movePosition(cursor.StartOfBlock)
        cursor.movePosition(cursor.Right, cursor.KeepAnchor, len(leadingWhitespace))

        # Compute the new indentation length, expanding any existing tabs
        indent = len(leadingWhitespace.expandtabs(self.indentWidth()))
        if self.indentUsingSpaces():
            # Determine correction, so we can round to multiples of indentation
            correction = indent % self.indentWidth()
            if correction and amount < 0:
                correction = -(self.indentWidth() - correction)  # Flip
            # Add the indentation tabs
            indent += (self.indentWidth() * amount) - correction
            cursor.insertText(" " * max(indent, 0))
        else:
            # Convert indentation to number of tabs, and add one
            indent = (indent // self.indentWidth()) + amount
            cursor.insertText("\t" * max(indent, 0))

    def dedentBlock(self, cursor):
        """dedentBlock(cursor)

        Dedent the block given by cursor.

        Calls indentBlock with amount = -1.
        May be overridden to customize indentation.

        """
        self.indentBlock(cursor, amount=-1)

    def indentSelection(self):
        """indentSelection()

        Called when the current line/selection is to be indented.
        Calls indentLine(cursor) for each line in the selection.
        May be overridden to customize indentation.

        See also doForSelectedBlocks and indentBlock.

        """
        self.doForSelectedBlocks(self.indentBlock)

    def dedentSelection(self):
        """dedentSelection()

        Called when the current line/selection is to be dedented.
        Calls dedentLine(cursor) for each line in the selection.
        May be overridden to customize indentation.

        See also doForSelectedBlocks and dedentBlock.

        """
        self.doForSelectedBlocks(self.dedentBlock)

    def justifyText(self, linewidth=70):
        """justifyText(linewidth=70)"""
        from .textutils import TextReshaper

        # Get cursor
        cursor = self.textCursor()

        # Make selection include whole lines
        pos1, pos2 = cursor.position(), cursor.anchor()
        pos1, pos2 = min(pos1, pos2), max(pos1, pos2)
        cursor.setPosition(pos1, cursor.MoveAnchor)
        cursor.movePosition(cursor.StartOfBlock, cursor.MoveAnchor)
        cursor.setPosition(pos2, cursor.KeepAnchor)
        cursor.movePosition(cursor.EndOfBlock, cursor.KeepAnchor)

        # Use reshaper to create replacement text
        reshaper = TextReshaper(linewidth)
        reshaper.pushText(cursor.selectedText())
        newText = reshaper.popText()

        # Update the selection
        # self.setTextCursor(cursor) for testing
        cursor.insertText(newText)

    def addLeftMargin(self, des, func):
        """Add a margin to the left. Specify a description for the margin,
        and a function to get that margin. For internal use.
        """
        assert des is not None
        self._leftmargins.append((des, func))

    def getLeftMargin(self, des=None):
        """Get the left margin, relative to the given description (which
        should be the same as given to addLeftMargin). If des is omitted
        or None, the full left margin is returned.
        """
        margin = 0
        for d, func in self._leftmargins:
            if d == des:
                break
            margin += func()
        return margin

    def updateMargins(self):
        """Force the margins to be recalculated and set the viewport
        accordingly.
        """
        leftmargin = self.getLeftMargin()
        self.setViewportMargins(leftmargin, 0, 0, 0)

    def toggleCase(self):
        """Change selected text to lower or upper case."""

        # Get cursor
        cursor = self.textCursor()
        # position = cursor.position()
        start_pos = cursor.selectionStart()
        end_pos = cursor.selectionEnd()

        # Get selected text
        selection = cursor.selectedText()

        if selection.islower():
            newText = selection.upper()
        elif selection.isupper():
            newText = selection.lower()
        else:
            newText = selection.lower()

        # Update the selection
        cursor.insertText(newText)
        cursor.setPosition(start_pos)
        cursor.setPosition(end_pos, QtGui.QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)



#
# Copyright © 2014-2015 Colin Duquesnoy
# Copyright © 2009- The Spyder Developmet Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides widget classes and functions.
"""
from . import PYQT5, PYQT6, PYSIDE2, PYSIDE6, PythonQtError


if PYQT6:
    from PyQt6 import QtWidgets
    from PyQt6.QtWidgets import *
    from PyQt6.QtGui import QAction, QActionGroup, QShortcut, QFileSystemModel
    from PyQt6.QtOpenGLWidgets import QOpenGLWidget

    # Map missing/renamed methods
    QTextEdit.setTabStopWidth = QTextEdit.setTabStopDistance
    QTextEdit.tabStopWidth = QTextEdit.tabStopDistance
    QTextEdit.print_ = QTextEdit.print
    QPlainTextEdit.setTabStopWidth = QPlainTextEdit.setTabStopDistance
    QPlainTextEdit.tabStopWidth = QPlainTextEdit.tabStopDistance
    QPlainTextEdit.print_ = QPlainTextEdit.print
    QApplication.exec_ = QApplication.exec
    QDialog.exec_ = QDialog.exec
    QMenu.exec_ = QMenu.exec

    # Allow unscoped access for enums inside the QtWidgets module
    from .enums_compat import promote_enums

    promote_enums(QtWidgets)
    del QtWidgets
elif PYQT5:
    from PyQt5.QtWidgets import *
elif PYSIDE6:
    from PySide6.QtWidgets import *
    from PySide6.QtGui import QAction, QActionGroup, QShortcut
    from PySide6.QtOpenGLWidgets import QOpenGLWidget

    # Map missing/renamed methods
    QTextEdit.setTabStopWidth = QTextEdit.setTabStopDistance
    QTextEdit.tabStopWidth = QTextEdit.tabStopDistance
    QPlainTextEdit.setTabStopWidth = QPlainTextEdit.setTabStopDistance
    QPlainTextEdit.tabStopWidth = QPlainTextEdit.tabStopDistance

    # Map DeprecationWarning methods
    QApplication.exec_ = QApplication.exec
    QDialog.exec_ = QDialog.exec
    QMenu.exec_ = QMenu.exec
elif PYSIDE2:
    from PySide2.QtWidgets import *
else:
    raise PythonQtError("No Qt bindings could be found")



#
# Copyright © 2009- The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides QtPrintSupport classes and functions.
"""

from . import PYQT5, PYQT6, PYSIDE6, PYSIDE2, PythonQtError


if PYQT5:
    from PyQt5.QtPrintSupport import *
elif PYQT6:
    from PyQt6.QtPrintSupport import *

    QPageSetupDialog.exec_ = QPageSetupDialog.exec
    QPrintDialog.exec_ = QPrintDialog.exec
    QPrintPreviewWidget.print_ = QPrintPreviewWidget.print
elif PYSIDE6:
    from PySide6.QtPrintSupport import *

    # Map DeprecationWarning methods
    QPageSetupDialog.exec_ = QPageSetupDialog.exec
    QPrintDialog.exec_ = QPrintDialog.exec
elif PYSIDE2:
    from PySide2.QtPrintSupport import *
else:
    raise PythonQtError("No Qt bindings could be found")



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

"""
Magic commands for the Pyzo kernel.
No need to use printDirect here, magic commands are just like normal Python
commands, in the sense that they print something etc.
"""

import sys
import os
import re
import time
import inspect
import tokenize, token, keyword
import itertools

try:
    import io
except ImportError:  # Not on Python 2.4
    pass

# Set Python version and get some names
PYTHON_VERSION = sys.version_info[0]
if PYTHON_VERSION < 3:
    input = raw_input  # noqa

MESSAGE = """List of *magic* commands:
    ?               - show this message
    ?X or X?        - show docstring of X
    ??X or X??      - help(X)
    cd              - show current directory
    cd X            - change directory
    ls              - list current directory
    who             - list variables in current workspace
    whos            - list variables plus their class and representation
    timeit X        - times execution of command X
    open X          - open file X or the Python module that defines object X
    run X           - run file X
    install         - install a new package
    update          - update an installed package
    remove          - remove (i.e. uninstall) an installed package
    uninstall       - alias for remove
    conda           - manage packages using conda
    pip             - manage packages using pip
    db X            - debug commands
    cls             - clear screen
    notebook        - launch the Jupyter notebook
"""


TIMEIT_MESSAGE = """Time execution duration. Usage:
    timeit fun  # where fun is a callable
    timeit 'expression'
    timeit expression
    timeit 20 fun/expression  # tests 20 passes
"""


def _detect_equalbang(line):
    try:
        gtoks = tokenize.tokenize(io.BytesIO(line.encode("utf-8")).readline)
        # iteration on two successive elements, https://docs.python.org/3/library/itertools.html#itertools-recipes
        first, second = itertools.tee(gtoks)
        next(second, None)
        for tok1, tok2 in zip(first, second):
            if (
                tok1.type == token.OP
                and tok1.string == "="
                and tok2.type == token.ERRORTOKEN
                and tok2.string == "!"
            ):
                return True
        return False
    except tokenize.TokenError:  # typically this means an unmatched parenthesis
        return False


def _should_not_interpret_as_magic(line):
    interpreter = sys._pyzoInterpreter

    # Check that line is not some valid python input
    try:
        # gets a list of 5-tuples, of which [0] is the type of token and [1] is the token string
        ltok = list(tokenize.tokenize(io.BytesIO(line.encode("utf-8")).readline))
    except tokenize.TokenError:
        # typically this means an unmatched parenthesis
        # (which should not happen because these are detected before)
        return True

    # ignore garbage and indentation at the beginning
    pos = 0
    while pos < len(ltok) and ltok[pos].type in [59, token.INDENT, tokenize.ENCODING]:
        # 59 is BACKQUOTE but there is no token.BACKQUOTE...
        pos = pos + 1
    # when line is only garbage or does not begin with a name
    if pos >= len(ltok) or ltok[pos].type != token.NAME:
        return True
    command = ltok[pos].string
    if keyword.iskeyword(command):
        return True
    pos = pos + 1
    # command is alone on the line
    if pos >= len(ltok) or ltok[pos].type in [token.ENDMARKER, tokenize.COMMENT]:
        if command in interpreter.locals:
            return True
        if interpreter.globals and command in interpreter.globals:
            return True
    else:  # command is not alone ; next token should not be an operator (this includes parentheses)
        if ltok[pos].type == token.OP and not line.endswith("?"):
            return True

    return False


class Magician:
    def _eval(self, command):

        # Get namespace
        NS1 = sys._pyzoInterpreter.locals
        NS2 = sys._pyzoInterpreter.globals
        if not NS2:
            NS = NS1
        else:
            NS = NS2.copy()
            NS.update(NS1)

        # Evaluate in namespace
        return eval(command, {}, NS)

    def convert_command(self, line):
        """convert_command(line)

        Convert a given command from a magic command to Python code.
        Returns the converted command if it was a magic command, or
        the original otherwise.

        """
        # Get converted command, catch and report errors
        try:
            res = self._convert_command(line)
        except Exception:
            # Try informing about line number
            type, value, tb = sys.exc_info()
            msg = "Error in handling magic function:\n"
            msg += "  %s\n" % str(value)
            if tb.tb_next:
                tb = tb.tb_next.tb_next
            while tb:
                msg += "  line %i in %s\n" % (
                    tb.tb_lineno,
                    tb.tb_frame.f_code.co_filename,
                )
                tb = tb.tb_next
            # Clear
            del tb
            # Write
            print(msg)
            return None

        # Process
        if res is None:
            return line
        else:
            return res

    def _convert_command(self, line):

        # Get interpreter
        interpreter = sys._pyzoInterpreter
        command = line.rstrip()

        if interpreter._ipython and _detect_equalbang(line):
            return self.equalbang_quirk(line, command)

        have_hard_chars = "cd ", "?"
        if PYTHON_VERSION >= 3 and not command.lower().startswith(have_hard_chars):
            try:
                if _should_not_interpret_as_magic(line):
                    return
            except Exception:
                pass  # dont break interpreter if above func has a bug ...

        else:
            # Old, not as good check for outdated Python version
            # Check if it is a variable
            if " " not in command:
                if command in interpreter.locals:
                    return
                if interpreter.globals and command in interpreter.globals:
                    return

        # Clean and make case insensitive
        command = line.upper().rstrip()

        # Commands to always support (also with IPython)

        if not command:
            return

        elif command.startswith("DB"):
            return self.debug(line, command)

        elif command.startswith("NOTEBOOK"):
            return self.notebook(line, command)

        elif command.startswith("INSTALL"):
            return self.install(line, command)

        elif command.startswith("UPDATE") or command.startswith("UPGRADE"):
            line = line.replace("upgrade", "update")
            command = command.replace("UPGRADE", "UPDATE")
            return self.update(line, command)

        elif command.startswith("REMOVE") or command.startswith("UNINSTALL"):
            line = line.replace("uninstall", "remove")
            return self.remove(line, command)

        elif command.startswith("CONDA"):
            return self.conda(line, command)

        elif command.startswith("PIP"):
            return self.pip(line, command)

        elif command == "CLS":
            return self.cls(line, command)

        elif command.startswith("OPEN "):
            return self.open(line, command)

        elif interpreter._ipython:
            # Patch IPython magic

            # EDIT do not run code after editing
            if command.startswith("EDIT ") and " -X " not in command:
                return "edit -x " + line[5:]

            # In all cases stop processing magic commands
            return

        # Commands that we only support in the absense of IPtython

        elif command == "?":
            return "print(%s)" % repr(MESSAGE)

        elif command.startswith("??"):
            return "help(%s)" % line[2:].rstrip()
        elif command.endswith("??"):
            return "help(%s)" % line.rstrip()[:-2]

        elif command.startswith("?"):
            return "print(%s.__doc__)" % line[1:].rstrip()

        elif command.endswith("?"):
            return "print(%s.__doc__)" % line.rstrip()[:-1]

        elif command.startswith("CD"):
            return self.cd(line, command)

        elif command.startswith("LS"):
            return self.ls(line, command)

        elif command.startswith("TIMEIT"):
            return self.timeit(line, command)

        elif command == "WHO":
            return self.who(line, command)

        elif command == "WHOS":
            return self.whos(line, command)

        elif command.startswith("RUN "):
            return self.run(line, command)

    def debug(self, line, command):
        if command == "DB":
            line = "db help"
        elif not command.startswith("DB "):
            return

        # Get interpreter
        interpreter = sys._pyzoInterpreter
        # Get command and arg
        line += " "
        _, cmd, arg = line.split(" ", 2)
        cmd = cmd.lower()
        # Get func
        func = getattr(interpreter.debugger, "do_" + cmd, None)
        # Call or show warning
        if func is not None:
            func(arg.strip())
        else:
            interpreter.write("Unknown debug command: %s.\n" % cmd)
        # Done (no code to execute)
        return ""

    def equalbang_quirk(self, line, command):
        print("`=!' is not valid Python. To check for non-equality, use `!='.")
        print(
            "If you intend to use IPython's ! magic feature and assign its result, please write `= !'."
        )
        return ""

    def cd(self, line, command):
        if command == "CD" or command.startswith("CD ") and "=" not in command:
            path = line[3:].strip()
            if path:
                try:
                    os.chdir(path)
                except Exception:
                    print('Could not change to directory "%s".' % path)
                    return ""
                newPath = os.getcwd()
            else:
                newPath = os.getcwd()
            # Done
            print(repr(newPath))
            return ""

    def ls(self, line, command):
        if command == "LS" or command.startswith("LS ") and "=" not in command:
            path = line[3:].strip()
            if not path:
                path = os.getcwd()
            L = [p for p in os.listdir(path) if not p.startswith(".")]
            text = "\n".join(sorted(L))
            # Done
            print(text)
            return ""

    def timeit(self, line, command):
        if command == "TIMEIT":
            return "print(%s)" % repr(TIMEIT_MESSAGE)
        elif command.startswith("TIMEIT "):
            expression = line[7:]
            # Get number of iterations
            N = 1
            tmp = expression.split(" ", 1)
            if len(tmp) == 2:
                try:
                    N = int(tmp[0])
                    expression = tmp[1]
                except Exception:
                    N = 1
            if expression[0] not in "'\"":
                isidentifier = lambda x: bool(re.match(r"[a-z_]\w*$", x, re.I))
                if not isidentifier(expression):
                    expression = "'%s'" % expression
            # Compile expression
            line2 = "import timeit; t=timeit.Timer(%s);" % expression
            line2 += "print(str( t.timeit(%i)/%i ) " % (N, N)
            line2 += '+" seconds on average for %i iterations." )' % N
            #
            return line2

    def who(self, line, command):
        L = self._eval("dir()\n")
        L = [k for k in L if not k.startswith("__")]
        if L:
            print(", ".join(L))
        else:
            print("There are no variables defined in this scope.")
        return ""

    def _justify(self, line, width, offset):
        realWidth = width - offset
        if len(line) > realWidth:
            line = line[: realWidth - 3] + "..."
        return line.ljust(width)

    def whos(self, line, command):
        # Get list of variables
        L = self._eval("dir()\n")
        L = [k for k in L if not k.startswith("__")]
        # Anny variables?
        if not L:
            print("There are no variables defined in this scope.")
            return ""
        else:
            text = "VARIABLE: ".ljust(20, " ") + "TYPE: ".ljust(20, " ")
            text += "REPRESENTATION: ".ljust(20, " ") + "\n"
        # Create list item for each variablee
        for name in L:
            ob = self._eval(name)
            cls = ""
            if hasattr(ob, "__class__"):
                cls = ob.__class__.__name__
            rep = repr(ob)
            text += self._justify(name, 20, 2) + self._justify(cls, 20, 2)
            text += self._justify(rep, 40, 2) + "\n"
        # Done
        print(text)
        return ""

    def cls(self, line, command):
        sys._pyzoInterpreter.context._strm_action.send("cls")
        return ""

    def open(self, line, command):

        # Get what to open
        name = line.split(" ", 1)[1].strip()
        fname = ""
        linenr = None

        # Is it a file name?
        tmp = os.path.join(os.getcwd(), name)
        #
        if name[0] in "\"'" and name[-1] in "\"'":  # Explicitly given
            fname = name[1:-1]
        elif os.path.isfile(tmp):
            fname = tmp
        elif os.path.isfile(name):
            fname = name

        else:
            # Then it maybe is an object

            # Get the object
            try:
                ob = self._eval(name)
            except NameError:
                print('There is no object known as "%s"' % name)
                return ""

            # Try get filename
            for iter in range(3):
                # Try successive steps
                if iter == 0:
                    ob = ob
                elif iter == 1 and not isinstance(ob, type):
                    ob = ob.__class__
                elif iter == 2 and hasattr(ob, "__module__"):
                    ob = sys.modules[ob.__module__]
                # Try get fname
                fname = ""
                try:
                    fname = inspect.getsourcefile(ob)
                except Exception:
                    pass
                # Returned fname may simply be x.replace('.pyc', '.py')
                if os.path.isfile(fname):
                    break

            # Try get line number
            if fname:
                try:
                    lines, linenr = inspect.getsourcelines(ob)
                except Exception:
                    pass

        # Almost done
        # IPython's edit now support this via our hook in interpreter.py
        if not fname:
            print('Could not determine file name for object "%s".' % name)
        elif linenr is not None:
            action = "open %i %s" % (linenr, os.path.abspath(fname))
            sys._pyzoInterpreter.context._strm_action.send(action)
        else:
            action = "open %s" % os.path.abspath(fname)
            sys._pyzoInterpreter.context._strm_action.send(action)
        #
        return ""

    def run(self, line, command):

        # Get what to open
        name = line.split(" ", 1)[1].strip()
        fname = ""

        # Enable dealing with qoutes
        if name[0] in "\"'" and name[-1] in "\"'":
            name = name[1:-1]

        # Is it a file name?
        tmp = os.path.join(os.getcwd(), name)
        #
        if os.path.isfile(tmp):
            fname = tmp
        elif os.path.isfile(name):
            fname = name

        # Go run!
        if not fname:
            print('Could not find file to run "%s".' % name)
        else:
            sys._pyzoInterpreter.runfile(fname)
        #
        return ""

    def _hasconda(self):
        try:
            from conda import __version__  # noqa
        except ImportError:
            return False
        return True

    def install(self, line, command):
        if not command.startswith("INSTALL "):
            return

        hasconda = self._hasconda()

        if hasconda:
            text = "Trying installation with conda. If this does not work, try:\n"
            text += "   pip " + line + "\n"
        else:
            text = "Trying installation with pip\n"

        print("\x1b[34m\x1b[1m" + text + "\x1b[0m")
        time.sleep(0.2)

        if hasconda:
            self.conda("conda " + line, "CONDA")
        else:
            self.pip("pip " + line, "PIP")
        return ""

    def update(self, line, command):
        if not command.startswith("UPDATE "):
            return

        hasconda = self._hasconda()

        if hasconda:
            text = "Trying update with conda. If this does not work, try:\n"
            text += "   pip " + line.replace("update", "install") + " --upgrade\n"
        else:
            text = "Trying pip install --upgrade\n"

        print("\x1b[34m\x1b[1m" + text + "\x1b[0m")
        time.sleep(0.2)

        if hasconda:
            self.conda("conda " + line, "CONDA")
        else:
            self.pip("pip " + line.replace("update", "install") + " --upgrade", "PIP")
        return ""

    def remove(self, line, command):
        if not command.startswith(("REMOVE ", "UNINSTALL")):
            return

        hasconda = self._hasconda()

        if hasconda:
            text = "Trying remove/uninstall with conda. If this does not work, try:\n"
            text += "   pip " + line.replace("remove", "uninstall") + "\n"
        else:
            text = "Trying remove/uninstall with pip\n"

        if hasconda:
            self.conda("conda " + line, "CONDA")
        else:
            self.pip("pip " + line.replace("remove", "uninstall"), "PIP")
        return ""

    def conda(self, line, command):

        if not (command == "CONDA" or command.startswith("CONDA ")):
            return

        # Get command args
        args = line.split(" ")
        args = [w for w in args if w]
        args.pop(0)  # remove 'conda'

        # Stop if user does "conda = ..."
        if args and "=" in args[0]:
            return

        # Add channels when using install, this gets added last, so
        # that user-specified channels take preference
        channel_list = []
        if args and args[0] == "install":
            channel_list = ["-c", "conda-forge", "-c", "pyzo"]

        def write_no_dots(x):
            if x.strip() == ".":  # note, no "x if y else z" in Python 2.4
                return 0
            return stderr_write(x)

        # Go!
        # Weird double-try, to make work on Python 2.4
        oldargs = sys.argv
        stderr_write = sys.stderr.write
        try:
            try:
                # older version of conda would spew dots to stderr during downloading
                sys.stderr.write = write_no_dots
                import conda  # noqa
                from conda.cli import main

                sys.argv = ["conda"] + list(args) + channel_list
                main()
            except SystemExit:  # as err:
                type, err, tb = sys.exc_info()
                del tb
                err = str(err)
                if len(err) > 4:  # Only print if looks like a message
                    print(err)
            except Exception:  # as err:
                type, err, tb = sys.exc_info()
                del tb
                print("Error in conda command:")
                print(err)
        finally:
            sys.argv = oldargs
            sys.stderr.write = stderr_write

        return ""

    # todo: I think we can deprecate this
    def _check_imported_modules(self):
        KNOWN_PURE_PYHON = (
            "conda",
            "yaml",
            "IPython",
            "requests",
            "readline",
            "pyreadline",
        )
        if not sys.platform.startswith("win"):
            return  # Only a problem on Windows
        # Check what modules are currently imported
        loaded_modules = set()
        for name, mod in sys.modules.items():
            if "site-packages" in getattr(mod, "__file__", ""):
                name = name.split(".")[0]
                if name.startswith("_") or name in KNOWN_PURE_PYHON:
                    continue
                loaded_modules.add(name)
        # Add PySide PyQt4 from IEP if prefix is the same
        if os.getenv("IEP_PREFIX", "") == sys.prefix:
            loaded_modules.add(os.getenv("IEP_QTLIB", "qt"))
        # Warn if we have any such modules
        loaded_modules = [m.lower() for m in loaded_modules if m]
        if loaded_modules:
            print("WARNING! The following modules are currently loaded:\n")
            print("  " + ", ".join(sorted(loaded_modules)))
            print(
                "\nUpdating or removing them may fail if they are not "
                "pure Python.\nIf none of the listed packages is updated or "
                'removed, it is safe\nto proceed (use "f" if necessary).\n'
            )
            print("-" * 80)
            time.sleep(2.0)  # Give user time to realize there is a warning

    def pip(self, line, command):

        if not (command == "PIP" or command.startswith("PIP ")):
            return

        # Get command args
        args = line.split(" ")
        args = [w for w in args if w]
        args.pop(0)  # remove 'pip'

        # Stop if user does "pip = ..."
        if args and "=" in args[0]:
            return

        # Tweak the args
        if args and args[0] == "uninstall":
            args.insert(1, "--yes")

        # Go!
        try:
            from pyzokernel.pipper import pip_command

            pip_command(*args)
        except SystemExit:  # as err:
            type, err, tb = sys.exc_info()
            del tb
            err = str(err)
            if len(err) > 4:  # Only print if looks like a message
                print(err)
        except Exception:  # as err:
            type, err, tb = sys.exc_info()
            del tb
            print("Error in pip command:")
            print(err)

        return ""

    def notebook(self, line, command):

        if not (command == "NOTEBOOK" or command.startswith("NOTEBOOK ")):
            return

        # Get command args
        args = line.split(" ")
        args = [w.replace("%20", " ") for w in args if w]
        args.pop(0)  # remove 'notebook'

        # Stop if user does "conda = ..."
        if args and "=" in args[0]:
            return

        # Go!
        # Weird double-try, to make work on Python 2.4
        oldargs = sys.argv
        try:
            try:
                import notebook.notebookapp

                sys.argv = ["jupyter_notebook"] + list(args)
                notebook.notebookapp.main()
            except SystemExit:  # as err:
                type, err, tb = sys.exc_info()
                del tb
                err = str(err)
                if len(err) > 4:  # Only print if looks like a message
                    print(err)
            except Exception:  # as err:
                type, err, tb = sys.exc_info()
                del tb
                print("Error in notebook command:")
                print(err)
        finally:
            sys.argv = oldargs

        return ""
